import os
import subprocess


def test_task_convert():
    subprocess.run(["yucca_convert_task", "-t", "Task000_TEST_SEGMENTATION"], check=True)
    subprocess.run(["yucca_convert_task", "-t", "Task000_TEST_CLASSIFICATION"], check=True)
    subprocess.run(["yucca_convert_task", "-t", "Task000_TEST_UNSUPERVISED"], check=True)


def test_preprocessing():
    subprocess.run(["yucca_preprocess", "-t", "Task000_TEST_SEGMENTATION"], check=True)
    subprocess.run(["yucca_preprocess", "-t", "Task000_TEST_UNSUPERVISED", "-pl", "UnsupervisedPlanner"], check=True)
    subprocess.run(
        [
            "yucca_preprocess",
            "-t",
            "Task000_TEST_CLASSIFICATION",
            "-pl",
            "YuccaPlanner_MaxSize",
            "-pr",
            "ClassificationPreprocessor",
        ],
        check=True,
    )


def test_training():
    # First: a very basic short training
    subprocess.run(
        [
            "yucca_train",
            "-t",
            "Task000_TEST_SEGMENTATION",
            "-m",
            "TinyUNet",
            "--epochs",
            "2",
            "--batch_size",
            "2",
            "--disable_logging",
            "--train_batches_per_step",
            "3",
            "--val_batches_per_step",
            "3",
        ],
        check=True,
    )

    # First: a very basic short training
    subprocess.run(
        [
            "yucca_train",
            "-t",
            "Task000_TEST_SEGMENTATION",
            "-m",
            "TinyUNet",
            "-man",
            "YuccaManager_AllAlways",
            "--epochs",
            "2",
            "--batch_size",
            "2",
            "--patch_size",
            "32",
            "--disable_logging",
            "--train_batches_per_step",
            "1",
            "--val_batches_per_step",
            "1",
        ],
        check=True,
    )
def test_finetune():
    from yucca.paths import yucca_models

    chk = os.path.join(
        yucca_models,
        "Task000_TEST_SEGMENTATION",
        "TinyUNet__3D",
        "YuccaManager__YuccaPlanner",
        "default",
        "kfold_5_fold_0",
        "version_0",
        "checkpoints",
        "best.ckpt",
    )
    # Second: a training with a lot of changed parameters
    subprocess.run(
        [
            "yucca_finetune",
            "-t",
            "Task000_TEST_UNSUPERVISED",
            "-chk",
            chk,
            "-m",
            "TinyUNet",
            "-d",
            "2D",
            "-pl",
            "UnsupervisedPlanner",
            "--lr",
            "0.0006",
            "--mom",
            "0.99",
            "--disable_logging",
            "--new_version",
            "--profile",
            "--experiment",
            "NonDefault",
            "--patch_size",
            "32",
            "--split_idx",
            "0",
            "--split_data_param",
            "0.7",
            "--split_data_method",
            "simple_train_val_split",
            "--epochs",
            "2",
            "--batch_size",
            "2",
            "--train_batches_per_step",
            "2",
            "--val_batches_per_step",
            "2",
        ],
        check=True,
    )


def test_inference():
    subprocess.run(
        [
            "yucca_inference",
            "-s",
            "Task000_TEST_SEGMENTATION",
            "-t",
            "Task000_TEST_SEGMENTATION",
            "-d",
            "3D",
            "-m",
            "TinyUNet",
            "--no_wandb",
        ],
        check=True,
    )


def test_evaluation():
    subprocess.run(
        [
            "yucca_evaluation",
            "-s",
            "Task000_TEST_SEGMENTATION",
            "-t",
            "Task000_TEST_SEGMENTATION",
            "-d",
            "3D",
            "-m",
            "TinyUNet",
            "--no_wandb",
            "--surface_eval",
        ],
        check=True,
    )

    subprocess.run(
        [
            "yucca_evaluation",
            "-s",
            "Task000_TEST_SEGMENTATION",
            "-t",
            "Task000_TEST_SEGMENTATION",
            "-d",
            "3D",
            "-m",
            "TinyUNet",
            "--no_wandb",
            "--obj_eval",
        ],
        check=True,
    )



from yucca.training.managers.YuccaManager import YuccaManager
from yucca.training.loss_and_optim.loss_functions.CE import CE


class YuccaManager_CE(YuccaManager):
    def __init__(
        self,
        model,
        model_dimensions: str,
        task: str,
        folds: str | int,
        plan_id: str,
        starting_lr: float = None,
        loss_fn: str = None,
        momentum: float = None,
        continue_training: bool = False,
    ):
        super().__init__(model, model_dimensions, task, folds, plan_id, starting_lr, loss_fn, momentum, continue_training)
        self._DEFAULT_LOSS = CE



from batchgenerators.transforms.abstract_transforms import AbstractTransform
from abc import abstractmethod


class YuccaTransform(AbstractTransform):
    @abstractmethod
    def get_params(self):
        """
        This will return a random value between
        values in a Tuple(low, high)
        e.g.
            sigma = (0, 10)
            val = np.random.uniform(*sigma)

        This is done for each random
        parameter of the augmentation.

        This method can be called
        (1) once for all samples of a batch
        (2) individually for each sample
        (3) individually for each modality of a sample

        This allows a high degree of flexibility.
        In label-preserving transforms we often use (1)
        to create a higher degree of variation
        In transforms that are not label-preserving we
        use either (2) or (3) with (2) being the
        preferred choice.
        This allows different intensities for each
        sample in a batch, while using the same
        intensity for all modalities of a sample
        and for both image and label.
        This is extremely important for transforms
        such as rotation, where we must ensure
        data modalities and labels remain registered.
        """

    @abstractmethod
    def __call__(self):
        """
        This will be of the form __call__(self, packed_dict: dict = None, **unpacked_dict):
        which allows calling it as either transform(data_dict) or transform(**data_dict),
        supporting both Torch pipelines and batchgenerators.
        """



from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.task_conversion.utils import generate_dataset_json, dirs_in_dir, should_use_volume
from yucca.paths import yucca_raw_data
from datetime import datetime
from tqdm import tqdm
import nibabel as nib


accept_modalities = ["t1", "t2", "mprage", "flair", "gre", "dwi", "swi", "grappa"]
rejec_modalities = [
    "resting_state",
    "fmri",
    "corrected",
    "mt",
    "repeat",
    "space",
]


def should_skip(modality: str):
    modality = modality.lower()
    if "2d" in modality:
        return True
    for reject in rejec_modalities:
        if reject in modality:
            return True
    for accept in accept_modalities:
        if accept in modality:
            return False
    return True


def convert(path: str, subdir: str = "PPMI"):
    """INPUT DATA - Define input path and suffixes"""
    path = join(path, subdir)

    """ OUTPUT DATA - Define the task name and prefix """
    task_name = "Task201_PPMI"
    task_prefix = "PPMI"

    subjects_dir = join(path, "DATA")
    target_base = join(yucca_raw_data, task_name)
    target_imagesTr = join(target_base, "imagesTr")

    maybe_mkdir_p(target_imagesTr)

    for subject in tqdm(dirs_in_dir(subjects_dir), desc="Subject"):
        subject_dir = join(subjects_dir, subject)
        for modality in dirs_in_dir(subject_dir):
            modality_dir = join(subject_dir, modality)
            if should_skip(modality):
                continue  # skip this modality
            for date in dirs_in_dir(modality_dir):
                date_dir = join(modality_dir, date)
                parsed_date = datetime.strptime(date, "%Y-%m-%d_%H_%M_%S.%f")
                date_simple = parsed_date.strftime("%Y%m%d")
                for image in dirs_in_dir(date_dir):
                    image_dir = join(date_dir, image)
                    if "nifti" and "dicom" in dirs_in_dir(image_dir):
                        image_dir = join(image_dir, "nifti")
                    for file in subfiles(image_dir, join=False, suffix=".nii"):
                        image_path = join(image_dir, file)
                        vol = nib.load(image_path)
                        if should_use_volume(vol):
                            output_name = f"{task_prefix}_{subject}_{modality}_{date_simple}_{image}_000.nii.gz"
                            output_path = join(target_imagesTr, output_name)
                            nib.save(vol, filename=output_path)

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        imagesTs_dir=None,
        modalities=["MRI"],
        labels=None,
        dataset_name=task_name,
        license="https://www.ppmi-info.org/sites/default/files/docs/ppmi-data-use-agreement.pdf",
        dataset_description="Parkinson Progression Markers Initiative",
        dataset_reference="https://www.ppmi-info.org/",
    )



from yucca.training.managers.YuccaManager import YuccaManager


class YuccaManager_PatchSizeMax(YuccaManager):
    """
    Instantiate a YuccaLightningManager object with the patch size set to the maximum patch size
    (no sliding window prediction).
    """

    def __init__(
        self,
        ckpt_path: str = None,
        continue_from_most_recent: bool = True,
        deep_supervision: bool = False,
        loss: str = "DiceCE",
        max_epochs: int = 1000,
        model_dimensions: str = "3D",
        model_name: str = "TinyUNet",
        num_workers: int = 8,
        planner: str = "YuccaPlanner",
        precision: str = "16-mixed",
        profile: bool = False,
        step_logging: bool = False,
        task: str = None,
        **kwargs,
    ):
        super().__init__(
            ckpt_path=ckpt_path,
            continue_from_most_recent=continue_from_most_recent,
            deep_supervision=deep_supervision,
            loss=loss,
            max_epochs=max_epochs,
            model_dimensions=model_dimensions,
            model_name=model_name,
            num_workers=num_workers,
            planner=planner,
            precision=precision,
            profile=profile,
            step_logging=step_logging,
            task=task,
            patch_size="max",
            patch_based_training=False,
            **kwargs,
        )



"""
Task_conversion file for the Fetal Planes DB (Barcelona) dataset (https://doi.org/10.5281/zenodo.3904280).

This is an example of a classification dataset with 2D PNG images and image-level labels.
"""

import shutil

import numpy as np
import pandas as pd
from tqdm import tqdm

from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p
from yucca.paths import yucca_raw_data
from yucca.task_conversion.utils import generate_dataset_json


def convert(path: str, subdir: str = "FETAL_PLANES_DB"):
    # INPUT DATA

    path = f"{path}/{subdir}"
    # file_suffix = ".png"

    # Train and Test images are in the same folder
    images_dir = join(path, "Images")

    # OUTPUT DATA
    # Target names
    task_name = "Task501_FetalPlanesDB"
    prefix = "FetalPlanesDB"

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    # collect labels
    data_df = pd.read_csv(join(path, "FETAL_PLANES_DB_data.csv"), delimiter=";")

    # Since there's trailing whitespaces in the columns of the dataset we strip names.
    data_df = data_df.rename(columns=lambda x: x.strip())
    labels = list(data_df["Plane"].unique())

    # Populate Target Directory
    # This is likely also the place to apply any re-orientation, resampling and/or label correction.
    for index, row in tqdm(data_df.iterrows(), total=len(data_df)):
        serial_number = index
        image_file_path = join(images_dir, f"{row['Image_name']}.png")
        label = np.array([int(labels.index(row["Plane"]))], dtype=np.int64)
        is_train_split = row["Train"] == 1
        if is_train_split:
            shutil.copyfile(image_file_path, f"{target_imagesTr}/{prefix}_{serial_number}_000.png")
            np.savetxt(f"{target_labelsTr}/{prefix}_{serial_number}.txt", label, fmt="%d")
        else:
            shutil.copyfile(image_file_path, f"{target_imagesTs}/{prefix}_{serial_number}_000.png")
            np.savetxt(f"{target_labelsTs}/{prefix}_{serial_number}.txt", label, fmt="%d")

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        ("US2D",),
        labels={i: l for i, l in enumerate(labels)},
        dataset_name=task_name,
        license="CC-BY 4.0",
        dataset_description="Fetal Planes DB (Barcelona)",
        dataset_reference="https://doi.org/10.5281/zenodo.3904280",
    )



import logging
from typing import Dict, List
from batchgenerators.utilities.file_and_folder_operations import subfiles, join, load_json, save_pickle
from yucca.utils.nib_utils import get_nib_spacing
from yucca.utils.type_conversions import nifti_or_np_to_np
from yucca.utils.loading import read_file_to_nifti_or_np
import nibabel as nib
import numpy as np
from tqdm.contrib.concurrent import process_map
from tqdm import tqdm
from functools import partial


def create_dataset_properties(data_dir, save_dir, suffix=".nii.gz", num_workers=12):
    """ "
    This will create a preprocessing-agnostic .pkl file containing basic dataset properties.
    We create this as a stand-alone file as the properties will be used by all preprocessors and should thus not
    be computed each time, as that will be very time consuming for memory intensive for larger datasets
    """
    dataset_json = load_json(join(data_dir, "dataset.json"))

    image_extension = dataset_json.get("image_extension") or suffix[1:]

    properties = {
        "image_extension": image_extension,
        "classes": list(dataset_json["labels"].keys()),
        "tasks": {task: [] for task in dataset_json["tasks"]},
        "label_hierarchy": dataset_json["label_hierarchy"],
        "modalities": dataset_json["modality"],
    }

    if len(dataset_json["tasks"]) > 0:
        assert not dataset_json["label_hierarchy"], "Multi Task implementation currently doesn't support Label Hierarchies"
        properties["classes"] = [list(dataset_json["labels"][task].keys()) for task in dataset_json["tasks"]]

    intensities = []
    sizes = []
    spacings = []

    modalities = properties["modalities"].items()
    mod_ids = list(list(zip(*modalities))[0])
    assert sorted(mod_ids) == mod_ids

    # Modalities are processed from smallest id to largest
    for mod_id, mod_name in modalities:
        mod_id = int(mod_id)
        suffix = f"_{mod_id:03}.{image_extension}"
        images_dir = join(data_dir, "imagesTr")

        subjects = subfiles(images_dir, suffix=suffix)
        if len(dataset_json["tasks"]) > 0:
            for task in dataset_json["tasks"]:
                for subject in subfiles(join(data_dir, "imagesTr", task), suffix=suffix, join=False):
                    if subject.endswith(suffix):
                        properties["tasks"][task].append(subject[: -len(suffix)])
                    subjects.append(join(data_dir, "imagesTr", task, subject))

        assert subjects, (
            f"no subjects found in {images_dir}. Ensure samples are "
            "suffixed with the modality encoding (such as '_000' for the first/only modality)"
            f"Looked for a files with ending {suffix}"
        )

        if "CT" in mod_name.upper():
            # -1000 == Air in HU (see https://en.wikipedia.org/wiki/Hounsfield_scale)
            # In practice it can be -1024, -1023 or in some cases -2048
            print("Using Hounsfield Units. Changing background pixel value from 0 to -1000")
            background_pixel_value = -1000
        else:
            background_pixel_value = 0

        chunksize = 50 if len(subjects) > 1000 else 1

        # `process_map` is a `tqdm` wrapper around `multiprocessing.Pool.map`
        map_result = process_map(
            partial(process, background_pixel_value=background_pixel_value),
            subjects,
            max_workers=num_workers,
            chunksize=chunksize,
            desc="Map",
        )  # returns list of dictionaries
        metadata = reduce(map_result)  # returns dictionary of metadata

        intensities.append(
            {
                "mean": float(np.mean(metadata["means"])),
                "min": float(np.min(metadata["mins"])),
                "max": float(np.max(metadata["maxs"])),
                "maxmin": float(np.max(metadata["mins"])),
                "std": float(np.mean(metadata["stds"])),
            }
        )

        # cross-modality metadata
        sizes = sizes + metadata["sizes"]
        spacings = spacings + metadata["spacings"]

    dims = [len(size) for size in sizes]
    assert all([dim == dims[0] for dim in dims]), f"not all volumes have the same number of dimensions. Sizes were: {dims}"

    properties["data_dimensions"] = int(len(sizes[0]))
    properties["original_sizes"] = sizes
    properties["original_spacings"] = spacings
    properties["intensities"] = intensities
    properties["original_max_size"] = np.max(sizes, 0).tolist()
    properties["original_min_size"] = np.min(sizes, 0).tolist()
    properties["original_median_size"] = np.median(sizes, 0).tolist()
    properties["original_max_spacing"] = np.max(spacings, 0).tolist()
    properties["original_min_spacing"] = np.min(spacings, 0).tolist()
    properties["original_median_spacing"] = np.median(spacings, 0).tolist()

    save_pickle(properties, join(save_dir, "dataset_properties.pkl"))


def reduce(results: List[Dict]):
    means = []
    mins = []
    maxs = []
    stds = []
    spacings = []
    sizes = []

    for res in tqdm(results, desc="Reduce"):
        means.append(res["mean"])
        mins.append(res["min"])
        maxs.append(res["max"])
        stds.append(res["std"])
        spacings.append(res["spacing"])
        sizes.append(res["size"])

    return {"means": means, "mins": mins, "maxs": maxs, "stds": stds, "spacings": spacings, "sizes": sizes}


def process(subject: str, background_pixel_value: int = 0):
    try:
        image = read_file_to_nifti_or_np(subject)
        dim = len(image.shape)

        if dim > 3:
            logging.warn(
                f"A volume has more than three dimensions. This is most often a mistake." f"Dims: {dim}, Vol: {subject}"
            )

        size = image.shape

        if isinstance(image, nib.Nifti1Image):
            spacing = get_nib_spacing(image).tolist()
        else:
            spacing = [1.0] * dim

        image = nifti_or_np_to_np(image)
        image_msk = image[image > background_pixel_value]

        mean = np.mean(image_msk)
        std = np.std(image_msk)

        min = np.min(image_msk)
        max = np.max(image_msk)

    except Exception as err:
        logging.warn(
            f"Could not read `{subject}`, got error `{err}`." "Suppressing to finalize, but you might need to act accordingly."
        )

    return {"size": size, "spacing": spacing, "min": min, "max": max, "mean": mean, "std": std}



import os
from typing import Literal
import numpy as np
import nibabel as nib
import json
import sys
import wandb
from batchgenerators.utilities.file_and_folder_operations import subfiles, join, load_json, isfile
from sklearn.metrics import confusion_matrix
from yucca.evaluation.metrics import (
    dice,
    jaccard,
    sensitivity,
    precision,
    TP,
    FP,
    FN,
    total_pos_gt,
    total_pos_pred,
    volume_similarity,
    accuracy,
    auroc,
)
from yucca.evaluation.obj_metrics import get_obj_stats_for_label
from yucca.evaluation.surface_metrics import get_surface_metrics_for_label
from yucca.paths import yucca_raw_data
from weave.monitoring import StreamTable
from tqdm import tqdm


class YuccaEvaluator(object):
    def __init__(
        self,
        labels: list | int,
        folder_with_predictions,
        folder_with_ground_truth,
        use_wandb: bool,
        as_binary=False,
        do_object_eval=False,
        do_surface_eval=False,
        overwrite: bool = False,
        task_type: Literal["segmentation", "classification", "regression"] = "segmentation",
    ):
        self.name = "results"

        self.overwrite = overwrite
        self.use_wandb = use_wandb
        self.task_type = task_type

        self.metrics = {
            "Dice": dice,
            "Jaccard": jaccard,
            "Sensitivity": sensitivity,
            "Precision": precision,
            "Volume Similarity": volume_similarity,
            "True Positives": TP,
            "False Positives": FP,
            "False Negatives": FN,
            "Total Positives Ground Truth": total_pos_gt,
            "Total Positives Prediction": total_pos_pred,
        }
        self.obj_metrics = []
        self.surface_metrics = []

        if self.task_type == "segmentation":
            self.metrics = {
                "Dice": dice,
                "Jaccard": jaccard,
                "Sensitivity": sensitivity,
                "Precision": precision,
                "Volume Similarity": volume_similarity,
                "True Positives": TP,
                "False Positives": FP,
                "False Negatives": FN,
                "Total Positives Ground Truth": total_pos_gt,
                "Total Positives Prediction": total_pos_pred,
            }

            if do_object_eval:
                self.name += "_OBJ"
                self.obj_metrics = [
                    "_OBJ Total Objects Prediction",
                    "_OBJ Total Objects Ground Truth",
                    "_OBJ True Positives",
                    "_OBJ False Positives",
                    "_OBJ False Negatives",
                    "_OBJ Mean Volume Prediction",
                    "_OBJ Mean Volume Ground Truth",
                    "_OBJ sensitivity",
                    "_OBJ precision",
                    "_OBJ F1",
                ]

            if do_surface_eval:
                self.name += "_SURFACE"
                self.surface_metrics = [
                    "Average Surface Distance",
                ]

            self.metrics_included_in_streamtable = [
                "Dice",
                "Sensitivity",
                "Precision",
                "Jaccard",
                "Volume Similarity",
                "_OBJ sensitivity",
                "_OBJ precision",
                "_OBJ F1",
            ]
        elif self.task_type == "classification":
            self.metrics = {
                "Accuracy": accuracy,
                "F1": dice,
                "Sensitivity": sensitivity,
                "Precision": precision,
                # "AUROC": auroc, # only when probabilities are available
            }

            self.metrics_included_in_streamtable = [
                "Accuracy",
                "F1",
                "Sensitivity",
                "Precision",
                # "AUROC", # only when probabilities are available
            ]
        elif self.task_type == "regression":
            raise NotImplementedError
            # self.metrics = {
            #     "MAE": mae,
            #     "RMSE": rmse,
            #     "R2": r2,
            # }
        else:
            raise ValueError(f"Unknown task type {self.task_type}")

        if isinstance(labels, int):
            self.labels = [str(i) for i in range(labels)]
        else:
            self.labels = labels
        self.as_binary = as_binary
        if self.as_binary:
            self.labels = ["0", "1"]
            self.name += "_BINARY"

        self.labelarr = np.sort(np.array(self.labels, dtype=np.uint8))
        self.folder_with_predictions = folder_with_predictions
        self.folder_with_ground_truth = folder_with_ground_truth

        self.outpath = join(self.folder_with_predictions, f"{self.name}.json")

        if self.task_type == "classification":
            self.pred_subjects = subfiles(self.folder_with_predictions, suffix=".txt", join=False)
            self.gt_subjects = subfiles(self.folder_with_ground_truth, suffix=".txt", join=False)
        else:
            self.pred_subjects = subfiles(self.folder_with_predictions, suffix=".nii.gz", join=False)
            self.gt_subjects = subfiles(self.folder_with_ground_truth, suffix=".nii.gz", join=False)

        print(
            f"\n"
            f"STARTING EVALUATION \n"
            f"Folder with predictions: {self.folder_with_predictions}\n"
            f"Folder with ground truth: {self.folder_with_ground_truth}\n"
            f"Evaluating performance on labels: {self.labels}"
        )

    def sanity_checks(self):
        assert self.pred_subjects <= self.gt_subjects, "Ground Truth is missing for predicted scans"

        assert self.gt_subjects <= self.pred_subjects, "Prediction is missing for Ground Truth of scans"

        # Check if the Ground Truth directory is a subdirectory of a 'TaskXXX_MyTask' folder.
        # If so, there should be a dataset.json where we can double check that the supplied classes
        # match with the expected classes for the dataset.
        gt_is_task = [i for i in self.folder_with_ground_truth.split(os.sep) if "Task" in i]
        if gt_is_task:
            gt_task = gt_is_task[0]
            dataset_json = join(yucca_raw_data, gt_task, "dataset.json")
            if isfile(dataset_json):
                dataset_json = load_json(dataset_json)
                print(f"Labels found in dataset.json: {list(dataset_json['labels'].keys())}")

    def run(self):
        if isfile(self.outpath) and not self.overwrite:
            print(f"Evaluation file already present in {self.outpath}. Skipping.")
        else:
            self.sanity_checks()
            results_dict = self.evaluate_folder()
            self.save_as_json(results_dict)
            if self.use_wandb:
                self.update_streamtable(results_dict)

    def evaluate_folder(self):
        if self.task_type == "classification":
            return self._evaluate_folder_cls()
        elif self.task_type == "segmentation":
            return self._evaluate_folder_segm()
        else:
            raise NotImplementedError("Invalid task type")

    def _evaluate_folder_cls(self):
        """
        Evaluate classification results
        """
        sys.stdout.flush()
        resultdict = {}

        predictions = []
        prediction_probs = []
        ground_truths = []

        # Flag to check if we have prediction probabilities to calculate AUROC
        use_probs = False

        # load predictions and ground truths
        for case in tqdm(self.pred_subjects, desc="Evaluating"):
            predpath = join(self.folder_with_predictions, case)
            gtpath = join(self.folder_with_ground_truth, case)

            pred: int = np.loadtxt(predpath)
            gt: int = np.loadtxt(gtpath)

            try:
                if len(prediction_probs) == 0:
                    print("Prediction probabilities found. Will use them for evaluation.")
                    use_probs = True

                pred_probs = np.load(predpath.replace(".txt", ".npz"))["data"]  # contains output probabilities
                prediction_probs.append(pred_probs)
            except FileNotFoundError:
                pred_probs = None

            predictions.append(pred)
            ground_truths.append(gt)

        predictions = np.array(predictions)
        ground_truths = np.array(ground_truths)

        if use_probs:
            prediction_probs = np.array(prediction_probs)
            assert len(predictions) == len(prediction_probs), (
                "Number of predicted labels and prediction probabilities do not match."
                "This likely means that the prediction probability file is missing for some of the predictions."
            )

            # add AUROC to streamtable metrics
            self.metrics_included_in_streamtable.append("AUROC")

        # calculate per-class metrics
        cmat = confusion_matrix(ground_truths, predictions, labels=self.labelarr)

        resultdict["per_class"] = {}

        for label in self.labelarr:
            tp = cmat[label, label]
            fp = sum(cmat[:, label]) - tp
            fn = sum(cmat[label, :]) - tp
            tn = np.sum(cmat) - tp - fp - fn

            labeldict = {}

            for k, v in self.metrics.items():
                labeldict[k] = round(v(tp, fp, tn, fn), 4)

            resultdict["per_class"][str(label)] = labeldict

        # calculate AUROC
        if use_probs:
            auroc_per_class: list[float] = auroc(ground_truths, prediction_probs)
            for label, score in zip(self.labelarr, auroc_per_class):
                resultdict["per_class"][str(label)]["AUROC"] = round(score, 4)

        # caclulate global (mean) metrics
        resultdict["mean"] = {}
        for k, _ in resultdict["per_class"][str(self.labelarr[0])].items():
            resultdict["mean"][k] = sum([resultdict["per_class"][str(label)][k] for label in self.labelarr])
            resultdict["mean"][k] = round(resultdict["mean"][k] / len(self.labelarr), 4)

        return resultdict

    def _evaluate_folder_segm(self):
        sys.stdout.flush()
        resultdict = {}
        meandict = {}

        for label in self.labels:
            meandict[label] = {k: [] for k in list(self.metrics.keys()) + self.obj_metrics + self.surface_metrics}

        for case in tqdm(self.pred_subjects, desc="Evaluating"):
            casedict = {}
            predpath = join(self.folder_with_predictions, case)
            gtpath = join(self.folder_with_ground_truth, case)

            pred = nib.load(predpath)
            gt = nib.load(gtpath)

            if self.as_binary:
                cmat = confusion_matrix(
                    np.around(gt.get_fdata().flatten()).astype(bool).astype(int),
                    np.around(pred.get_fdata().flatten()).astype(bool).astype(int),
                    labels=self.labelarr,
                )
            else:
                cmat = confusion_matrix(
                    np.around(gt.get_fdata().flatten()).astype(int),
                    np.around(pred.get_fdata().flatten()).astype(int),
                    labels=self.labelarr,
                )

            for label in self.labelarr:
                labeldict = {}

                tp = cmat[label, label]
                fp = sum(cmat[:, label]) - tp
                fn = sum(cmat[label, :]) - tp
                tn = np.sum(cmat) - tp - fp - fn  # often a redundant and meaningless metric
                for k, v in self.metrics.items():
                    labeldict[k] = round(v(tp, fp, tn, fn), 4)
                    meandict[str(label)][k].append(labeldict[k])

                if self.obj_metrics:
                    # now for the object metrics
                    obj_labeldict = get_obj_stats_for_label(gt, pred, label, as_binary=self.as_binary)
                    for k, v in obj_labeldict.items():
                        labeldict[k] = round(v, 4)
                        meandict[str(label)][k].append(labeldict[k])

                if self.surface_metrics:
                    surface_labeldict = get_surface_metrics_for_label(gt, pred, label, as_binary=self.as_binary)
                    for k, v in surface_labeldict.items():
                        labeldict[k] = round(v, 4)
                        meandict[str(label)][k].append(labeldict[k])
                casedict[str(label)] = labeldict
            casedict["Prediction:"] = predpath
            casedict["Ground Truth:"] = gtpath

            resultdict[case] = casedict
        for label in self.labels:
            meandict[label] = {
                k: round(np.nanmean(v), 4) if not np.all(np.isnan(v)) else 0 for k, v in meandict[label].items()
            }
        resultdict["mean"] = meandict

        return resultdict

    def save_as_json(self, dict):
        print(f"Saving results.json" "\n" "\n" f"########################################################################")
        with open(self.outpath, "w") as f:
            json.dump(dict, f, default=float, indent=4)

    def update_streamtable(self, results_dict):
        """
        Save evaluation results to a wandb StreamTable

        :param results_dict: dictionary with evaluation results
        """
        task = self.outpath.split(os.path.sep)[-5]
        target = self.outpath.split(os.path.sep)[-6]
        model_name = "/".join(self.outpath.split(os.path.sep)[-4:])

        st = StreamTable(table_name=task, entity_name=wandb.api.viewer()["entity"], project_name="Yucca")

        stream_dict = {"0. Experiment": model_name, "0. Target Task": target}

        if self.task_type == "classification":
            stream_dict = {**stream_dict, **results_dict}

        elif self.task_type == "segmentation":
            for key, _ in results_dict.items():
                if key == "0":
                    continue
                else:
                    stream_dict.update(
                        {f"{key}. " + k: v for k, v in results_dict[key].items() if k in self.metrics_included_in_streamtable}
                    )

        else:
            raise NotImplementedError("Task type not supported")

        st.log(stream_dict)
        st.finish()



import argparse
from yucca.paths import yucca_source
import importlib
import re


def remove_task_prefix(str):
    """
    Removes the "TaskXXX_" prefix from string. Number is arbitrary, and T need not be capitalized.
    Example:
      Input: "Task001_OASIS_IS_COOL"
      Output: "OASIS_IS_COOL"
    """
    pattern = re.compile(r"^Task\d+_", re.IGNORECASE)
    return pattern.sub("", str)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-t", "--task", help="Name of the task to preprocess. " "Should be of format: TaskXXX_MYTASK", required=True
    )
    parser.add_argument("-p", "--path", help="Path to source data", default=yucca_source)
    parser.add_argument("-d", "--subdir", help="Directory of data inside source data")

    args = parser.parse_args()
    task_converter = importlib.import_module(f"yucca.task_conversion.{args.task}")

    if args.subdir is None:
        task_converter.convert(args.path)
    else:
        task_converter.convert(args.path, args.subdir)


if __name__ == "__main__":
    main()



from yucca.planning.YuccaPlanner import YuccaPlanner


class YuccaPlanner_NoResample(YuccaPlanner):
    def __init__(self, task, view=None):
        super().__init__(task=task, view=view)
        self.name = str(self.__class__.__name__) + str(view or "")

    def determine_target_size_from_fixed_size_or_spacing(self):
        self.fixed_target_spacing = []
        self.fixed_target_size = []



from yucca.training.managers.YuccaManager import YuccaManager
from yucca.training.augmentation.augmentation_presets import all_always


class YuccaManager_AllAlways(YuccaManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.augmentation_params = all_always



import numpy as np
import torch
import torch.nn.functional as F
import nibabel as nib
import os
import cc3d
import logging
import math
import time
import re
from yucca.utils.loading import load_yaml, read_file_to_nifti_or_np
from yucca.image_processing.objects.BoundingBox import get_bbox_for_foreground
from yucca.image_processing.cropping_and_padding import crop_to_box, pad_to_size, get_pad_kwargs
from yucca.utils.nib_utils import (
    get_nib_spacing,
    get_nib_orientation,
    reorient_nib_image,
)
from yucca.utils.type_conversions import nifti_or_np_to_np
from yucca.paths import yucca_preprocessed_data, yucca_raw_data
from yucca.preprocessing.normalization import normalizer
from multiprocessing import Pool
from skimage.transform import resize
from batchgenerators.utilities.file_and_folder_operations import (
    join,
    load_json,
    subfiles,
    save_pickle,
    maybe_mkdir_p,
    isfile,
)


class YuccaPreprocessor(object):
    """
    The YuccaPreprocessor class is designed to preprocess medical images for the Yucca project.
    It implements various preprocessing steps, such as reorientation, cropping, normalization, and resizing,
    based on the plans specified in an YuccaPlanner.

    For training the _preprocess_train_subject method prepares input images for the Yucca model.
    The preprocess_case_for_inference method prepares input images for the Yucca model during the inference phase,
    ensuring that they match the requirements specified during training.
    The reverse_preprocessing method is then used to revert the processed images back to their original form,
    allowing for a meaningful interpretation of the model's predictions.
    These methods collectively provide a consistent and reversible preprocessing pipeline for both training and inference.

    The operations that can be enabled/defined in the YuccaPlanner and carried out by the
    YuccaPreprocessor are:

    (1) The starting orientation - defaults to RAS (for medical images).
    (2) The cropping operation - defaults to crop to nonzero bounding box
    (3) The Transposition operation (along with the reverse transpose operation,
    to be used during inference) - defaults to no transposition if image dimensions and spacings
    are not too anisotropic.
    (4) The Resample operation - defaults to resampling to the median spacing of the dataset.
    (5) The Normalization operation - defaults to standardization = (image - mean) / std
    per modality to preserve ranges to account for CT pixel values representing specific physical
    attributes.

    Additionally it carries out a number of tests and analyzes each image for foreground locations
    which is used later to oversample foreground.
    """

    def __init__(self, plans_path, task=None, threads=None, disable_sanity_checks=False, enable_cc_analysis=False):
        self.name = str(self.__class__.__name__)
        self.task = task
        self.plans_path = plans_path
        self.plans = self.load_plans(plans_path)
        self.threads = threads
        self.disable_sanity_checks = disable_sanity_checks
        self.enable_cc_analysis = enable_cc_analysis

        # lists for information we would like to attain
        self.transpose_forward = []
        self.transpose_backward = []
        self.target_spacing = []

    def initialize_paths(self):
        self.target_dir = join(yucca_preprocessed_data, self.task, self.plans["plans_name"])
        self.input_dir = join(yucca_raw_data, self.task)
        self.imagepaths = subfiles(join(self.input_dir, "imagesTr"), suffix=self.image_extension)
        self.subject_ids = [
            file for file in subfiles(join(self.input_dir, "labelsTr"), join=False) if not file.startswith(".")
        ]

    def initialize_properties(self):
        """
        here we basically set up things that are needed for preprocessing during training,
        but that aren't necessary during inference
        """
        self.dataset_properties = self.plans["dataset_properties"]
        self.intensities = self.dataset_properties["intensities"]
        self.image_extension = self.dataset_properties.get("image_extension") or "nii.gz"

        # op values
        self.transpose_forward = np.array(self.plans["transpose_forward"], dtype=int)
        self.transpose_backward = np.array(self.plans["transpose_backward"], dtype=int)
        self.target_spacing = self.plans["target_spacing"]
        self.target_size = (
            np.array(self.plans.get("target_size"), dtype=int) if self.plans.get("target_size") not in ["null", None] else None
        )

    def run(self):
        self.initialize_properties()
        self.initialize_paths()
        maybe_mkdir_p(self.target_dir)

        logging.info(
            f"{'Preprocessing Task:':25.25} {self.task} \n"
            f"{'Using Planner:':25.25} {self.plans_path} \n"
            f"{'Crop to nonzero:':25.25} {self.plans['crop_to_nonzero']} \n"
            f"{'Normalization scheme:':25.25} {self.plans['normalization_scheme']} \n"
            f"{'Transpose Forward:':25.25} {self.transpose_forward} \n"
            f"{'Transpose Backward:':25.25} {self.transpose_backward} \n"
            f"{'Number of threads:':25.25} {self.threads}"
        )
        p = Pool(self.threads)
        p.map_async(self.preprocess_train_subject, self.subject_ids)
        p.close()
        p.join()

    def preprocess_train_subject(self, subject_id):
        """
        This is the bread and butter of the preprocessor.
        The following steps are taken:

        (1) Load Images:
        Extract relevant image files associated with the given subject_id.
        Load the images using the nibabel library.

        (2) Reorientation (Optional):
        Check if valid qform or sform codes are present in the header.
        If valid, reorient the images to the target orientation specified in the plans.
        Update the original and new orientation information in the image_props dictionary.

        (3) Normalization and Transposition:
        Normalize each image based on the specified normalization scheme and intensities.
        Transpose the images according to the forward transpose axes specified in the plans.

        (4) Cropping (Optional):
        If the crop_to_nonzero option is enabled in the plans, crop the images to the nonzero bounding box.
        Update the image_props dictionary with cropping information.

        (5) Resampling:
        Resample images to the target spacing specified in the plans.
        Update the image_props dictionary with original and new spacing information.

        (6) Foreground Locations:
        Extract some locations of the foreground, which will be used in oversampling of foreground classes.
        Determine the number and sizes of connected components in the ground truth label (can be used in analysis).

        (7) Save Preprocessed Data:
        Stack the preprocessed images and label.
        Save the preprocessed data as a NumPy array in a .npy file.
        Save relevant metadata as a .pkl file.

        (8) Print Information:
        Print information about the size and spacing before and after preprocessing.
        Print the path where the preprocessed data is saved.
        """
        subject_id = subject_id.split(os.extsep, 1)[0]
        arraypath = join(self.target_dir, subject_id + ".npy")
        picklepath = join(self.target_dir, subject_id + ".pkl")

        if isfile(arraypath) and isfile(picklepath):
            logging.info(f"Case: {subject_id} already exists. Skipping.")
            return

        start_time = time.time()

        images, label, image_props = self._preprocess_train_subject(subject_id, label_exists=True, preprocess_label=True)

        # Stack and fix dimensions
        images = np.vstack((np.array(images), np.array(label)[np.newaxis]))

        # save the image
        np.save(arraypath, images)

        # save metadata as .pkl
        save_pickle(image_props, picklepath)

        end_time = time.time()
        logging.info(
            f"Preprocessed case: {subject_id} \n"
            f"size before: {image_props['original_size']} \n"
            f"cropping enabled: {self.plans['crop_to_nonzero']} \n"
            f"size after crop: {image_props['size_before_transpose']} \n"
            f"size final: {image_props['new_size']} \n"
            f"spacing before: {image_props['original_spacing']} spacing after: {image_props['new_spacing']} \n"
            f"Saving {subject_id} in {arraypath} \n"
            f"Time elapsed: {round(end_time-start_time, 4)} \n"
        )

    def _preprocess_train_subject(self, subject_id, label_exists: bool, preprocess_label: bool):
        image_props = {}

        # First find relevant images by their paths and save them in the image property pickle
        # Then load them as images
        # The '_' in the end is to avoid treating Case_4_000 AND Case_42_000 as different versions
        # of the label named Case_4 as both would start with "Case_4", however only the correct one is
        # followed by an underscore
        escaped_subject_id = re.escape(subject_id)

        # path to all modalities of subject_id
        imagepaths = [
            impath
            for impath in self.imagepaths
            # Check if impath is a modality of subject_id (subject_id + _XXX + .) where XXX are three digits
            if re.search(escaped_subject_id + "_" + r"\d{3}" + ".", os.path.split(impath)[-1])
        ]
        assert len(imagepaths) > 0, "found no images"
        image_props["image files"] = imagepaths
        images = [read_file_to_nifti_or_np(image) for image in imagepaths]
        if label_exists:
            # Do the same with label
            label = [
                labelpath
                for labelpath in subfiles(join(self.input_dir, "labelsTr"))
                if os.path.split(labelpath)[-1].startswith(subject_id + ".")
            ]
            assert len(label) == 1, f"unexpected number of labels found. Expected 1 and found {len(label)}"
            image_props["label file"] = label[0]
            label = read_file_to_nifti_or_np(label[0], dtype=np.uint8)
        else:
            label = None

        if not self.disable_sanity_checks:
            if label_exists and preprocess_label:
                self.run_sanity_checks(images, label, subject_id, imagepaths)
            else:
                self.run_sanity_checks(images, None, subject_id, imagepaths)

        images, label, image_props["nifti_metadata"] = self.apply_nifti_preprocessing_and_return_numpy(
            images, np.array(images[0].shape), label, include_header=False
        )

        original_size = images[0].shape

        if label_exists and preprocess_label:
            self.verify_label_validity(label, subject_id)

        # Cropping is performed to save computational resources. We are only removing background.
        if self.plans["crop_to_nonzero"]:
            nonzero_box = get_bbox_for_foreground(images[0], background_label=0)
            image_props["crop_to_nonzero"] = nonzero_box
            for i in range(len(images)):
                images[i] = crop_to_box(images[i], nonzero_box)
            if label_exists and preprocess_label:
                label = crop_to_box(label, nonzero_box)
        else:
            image_props["crop_to_nonzero"] = self.plans["crop_to_nonzero"]

        image_props["size_before_transpose"] = list(images[0].shape)
        if label_exists and preprocess_label:
            images, label = self.transpose_case(images, self.transpose_forward, label)
        else:
            images = self.transpose_case(images, self.transpose_forward, None)
        image_props["size_after_transpose"] = list(images[0].shape)
        resample_target_size, final_target_size, new_spacing = self.determine_target_size(
            images_transposed=images,
            original_spacing=np.array(image_props["nifti_metadata"]["original_spacing"]),
            transpose_forward=self.transpose_forward,
        )

        if label_exists and preprocess_label:
            images, label = self.resample_and_normalize_case(
                images=images,
                target_size=resample_target_size,
                label=label,
                norm_op=self.plans["normalization_scheme"],
            )
        else:
            images = self.resample_and_normalize_case(
                images=images,
                target_size=resample_target_size,
                label=None,
                norm_op=self.plans["normalization_scheme"],
            )

        if final_target_size is not None:
            if label_exists and preprocess_label:
                images, label = self.pad_to_size(images, size=final_target_size, label=label)
            else:
                images = self.pad_to_size(images, size=final_target_size, label=None)

        if label_exists and preprocess_label:
            image_props["foreground_locations"], image_props["label_cc_n"], image_props["label_cc_sizes"] = self.analyze_label(
                label=images[-1]
            )
        else:
            image_props["label_cc_n"] = image_props["label_cc_sizes"] = 0
            image_props["foreground_locations"] = []

        image_props["new_size"] = list(images[0].shape)

        # save relevant values
        image_props["original_spacing"] = image_props["nifti_metadata"]["original_spacing"]
        image_props["original_size"] = original_size
        image_props["original_orientation"] = image_props["nifti_metadata"]["original_orientation"]
        image_props["new_spacing"] = new_spacing
        image_props["new_direction"] = image_props["nifti_metadata"]["final_direction"]

        return images, label, image_props

    def preprocess_case_for_inference(self, images: list | tuple, patch_size: tuple, sliding_window_prediction: bool = True):
        """
        Will reorient ONLY if we have valid qform or sform codes.
        with coded=True the methods will return {affine or None} and {0 or 1}.
        If both are 0 we cannot rely on headers for orientations and will
        instead assume images are in the desired orientation already.

        Afterwards images will be normalized and transposed as specified by the
        plans file also used in training.

        Finally images are resampled to the required spacing/size and returned
        as torch tensors of the required shape (b, c, x, y, (z))
        """
        assert isinstance(images, (list, tuple)), "image(s) should be a list or tuple, even if only one " "image is passed"
        self.initialize_properties()

        if sliding_window_prediction is False:
            self.target_size = patch_size

        image_properties = {}
        ext = images[0][0].split(os.extsep, 1)[1] if isinstance(images[0], tuple) else images[0].split(os.extsep, 1)[1]
        images = [
            read_file_to_nifti_or_np(image[0]) if isinstance(image, tuple) else read_file_to_nifti_or_np(image)
            for image in images
        ]

        image_properties["image_extension"] = ext
        image_properties["original_shape"] = np.array(images[0].shape)

        assert len(image_properties["original_shape"]) in [
            2,
            3,
        ], "images must be either 2D or 3D for preprocessing"

        images, _, image_properties["nifti_metadata"] = self.apply_nifti_preprocessing_and_return_numpy(
            images, image_properties["original_shape"], label=None, include_header=True
        )

        image_properties["uncropped_shape"] = np.array(images[0].shape)

        if self.plans["crop_to_nonzero"]:
            nonzero_box = get_bbox_for_foreground(images[0], background_label=0)
            for i in range(len(images)):
                images[i] = crop_to_box(images[i], nonzero_box)
            image_properties["nonzero_box"] = nonzero_box

        image_properties["cropped_shape"] = np.array(images[0].shape)

        images = self.transpose_case(images, self.transpose_forward, None)

        resample_target_size, _, _ = self.determine_target_size(
            images_transposed=images,
            original_spacing=np.array(image_properties["nifti_metadata"]["original_spacing"]),
            transpose_forward=self.transpose_forward,
        )

        images = self.resample_and_normalize_case(
            images=images,
            target_size=resample_target_size,
            label=None,
            norm_op=self.plans["normalization_scheme"],
        )

        # From this point images are shape (1, c, x, y, z)
        image_properties["resampled_transposed_shape"] = np.array(images[0].shape)

        for i in range(len(images)):
            images[i], padding = pad_to_size(images[i], patch_size)
        image_properties["padded_shape"] = np.array(images[0].shape)
        image_properties["padding"] = padding

        # Stack and fix dimensions
        images = np.stack(images)[np.newaxis]

        return torch.tensor(images, dtype=torch.float32), image_properties

    def reverse_preprocessing(self, images: torch.Tensor, image_properties: dict):
        """
        Expected shape of images are:
        (b, c, x, y(, z))

        (1) Initialization: Extract relevant properties from the image_properties dictionary.
        (2) Padding Reversion: Reverse the padding applied during preprocessing.
        (3) Resampling and Transposition Reversion: Resize the images to revert the resampling operation.
        Transpose the images back to the original orientation.
        (4) Cropping Reversion (Optional): If cropping to the nonzero bounding box was applied, revert the cropping operation.
        (5) Return: Return the reverted images as a NumPy array.
        The original orientation of the image will be re-applied when saving the prediction
        """
        image_properties["save_format"] = image_properties.get("image_extension")
        nclasses = max(1, len(self.plans["dataset_properties"]["classes"]))
        canvas = torch.zeros((1, nclasses, *image_properties["uncropped_shape"]), dtype=images.dtype)
        shape_after_crop = image_properties["cropped_shape"]
        shape_after_crop_transposed = shape_after_crop[self.transpose_forward]
        pad = image_properties["padding"]

        assert np.all(images.shape[2:] == image_properties["padded_shape"]), (
            f"Reversing padding: "
            f"image should be of shape: {image_properties['padded_shape']}"
            f"but is: {images.shape[2:]}"
        )
        shape = images.shape[2:]
        if len(pad) == 6:
            images = images[
                :,
                :,
                pad[0] : shape[0] - pad[1],
                pad[2] : shape[1] - pad[3],
                pad[4] : shape[2] - pad[5],
            ]
        elif len(pad) == 4:
            images = images[:, :, pad[0] : shape[0] - pad[1], pad[2] : shape[1] - pad[3]]

        assert np.all(images.shape[2:] == image_properties["resampled_transposed_shape"]), (
            f"Reversing resampling and tranposition: "
            f"image should be of shape: {image_properties['resampled_transposed_shape']}"
            f"but is: {images.shape[2:]}"
        )
        # Here we Interpolate the array to the original size. The shape starts as [H, W (,D)]. For Torch functionality it is changed to [B, C, H, W (,D)].
        # Afterwards it's squeezed back into [H, W (,D)] and transposed to the original direction.
        images = F.interpolate(images, size=shape_after_crop_transposed.tolist(), mode="trilinear").permute(
            [0, 1] + [i + 2 for i in self.transpose_backward]
        )

        # Now move the tensor to the CPU
        images = images.cpu()
        assert np.all(images.shape[2:] == image_properties["cropped_shape"]), (
            f"Reversing cropping: "
            f"image should be of shape: {image_properties['cropped_shape']}"
            f"but is: {images.shape[2:]}"
        )

        if self.plans["crop_to_nonzero"]:
            bbox = image_properties["nonzero_box"]
            slices = [
                slice(None),
                slice(None),
                slice(bbox[0], bbox[1]),
                slice(bbox[2], bbox[3]),
            ]
            if len(bbox) == 6:
                slices.append(
                    slice(bbox[4], bbox[5]),
                )
            canvas[slices] = images
        else:
            canvas = images
        return canvas.numpy(), image_properties

    @staticmethod
    def load_plans(plans_path):
        if os.path.splitext(plans_path)[-1] == ".json":
            return load_json(plans_path)
        if os.path.splitext(plans_path)[-1] == ".yaml":
            return load_yaml(plans_path)["config"]["plans"]
        else:
            raise FileNotFoundError(
                f"Plan file not found. Got {plans_path} with ext {os.path.splitext(plans_path)[-1]}. Expects either a '.json' or '.yaml' file."
            )

    @staticmethod
    def pad_to_size(images: list, size, pad_value="min", label=None):
        for i in range(len(images)):
            pad_kwargs = get_pad_kwargs(data=images[i], pad_value=pad_value)
            images[i], _ = pad_to_size(images[i], size, **pad_kwargs)
        if label is None:
            return images
        label, _ = pad_to_size(label, size, **pad_kwargs)
        return images, label

    @staticmethod
    def transpose_case(
        images: list,
        transpose: np.ndarray,
        label=None,
    ):
        for i in range(len(images)):
            assert len(images[i].shape) == len(transpose), (
                "image and transpose axes do not match. \n"
                f"images[i].shape == {images[i].shape} \n"
                f"transpose == {transpose} \n"
                f"len(images[i].shape) == {len(images[i]).shape} \n"
                f"len(transpose) == {len(transpose)} \n"
            )
            images[i] = images[i].transpose(transpose)

        if label is not None:
            label = label.transpose(transpose)
            return images, label

        return images

    def analyze_label(self, label):
        # we get some (no need to get all) locations of foreground, that we will later use in the
        # oversampling of foreground classes
        # And we also potentially analyze the connected components of the label
        max_foreground_locs = 100000  # limited to save space
        foreground_locs = np.array(np.nonzero(label)).T[::10].tolist()
        if len(foreground_locs) > max_foreground_locs:
            foreground_locs = foreground_locs[:: round(len(foreground_locs) / max_foreground_locs)]
        if not self.enable_cc_analysis:
            label_cc_n = 0
            label_cc_sizes = 0
        else:
            numbered_ground_truth, label_cc_n = cc3d.connected_components(label, connectivity=26, return_N=True)
            if len(numbered_ground_truth) == 0:
                label_cc_sizes = 0
            else:
                label_cc_sizes = [
                    i * np.prod(self.target_spacing) for i in np.unique(numbered_ground_truth, return_counts=True)[-1][1:]
                ]
        return foreground_locs, label_cc_n, label_cc_sizes

    def apply_nifti_preprocessing_and_return_numpy(
        self,
        images,
        original_size,
        label=None,
        include_header=False,
    ):
        # If qform and sform are both missing the header is corrupt and we do not trust the
        # direction from the affine
        # Make sure you know what you're doing
        metadata = {
            "original_spacing": np.array([1.0] * len(original_size)).tolist(),
            "original_orientation": None,
            "final_direction": None,
            "header": None,
            "affine": None,
            "reoriented": False,
        }

        if isinstance(images[0], nib.Nifti1Image):
            # If qform and sform are both missing the header is corrupt and we do not trust the
            # direction from the affine
            # Make sure you know what you're doing
            if images[0].get_qform(coded=True)[1] or images[0].get_sform(coded=True)[1]:
                metadata["reoriented"] = True
                metadata["original_orientation"] = get_nib_orientation(images[0])
                metadata["final_direction"] = self.plans["target_coordinate_system"]
                images = [
                    reorient_nib_image(image, metadata["original_orientation"], metadata["final_direction"])
                    for image in images
                ]
                if label is not None and isinstance(label, nib.Nifti1Image):
                    label = reorient_nib_image(label, metadata["original_orientation"], metadata["final_direction"])
            if include_header:
                metadata["header"] = images[0].header
            metadata["original_spacing"] = get_nib_spacing(images[0]).tolist()
            metadata["affine"] = images[0].affine

        images = [nifti_or_np_to_np(image) for image in images]
        if label is not None:
            label = nifti_or_np_to_np(label)
        return images, label, metadata

    def determine_target_size(
        self,
        images_transposed: list,
        original_spacing,
        transpose_forward,
    ):
        final_target_size = None
        image_shape_t = np.array(images_transposed[0].shape)
        # We do not want to change the aspect ratio so we resample using the minimum alpha required
        # to attain 1 correct dimension, and then the rest will be padded.
        # Additionally we make sure each dimension is divisible by 16 to avoid issues with standard pooling/stride settings
        if self.target_size is not None:
            if self.plans["keep_aspect_ratio_when_using_target_size"] is True:
                resample_target_size = np.array(image_shape_t * np.min(self.target_size / image_shape_t)).astype(int)
                final_target_size = self.target_size
                final_target_size = [math.ceil(i / 16) * 16 for i in final_target_size]
            else:
                resample_target_size = self.target_size
                resample_target_size = [math.ceil(i / 16) * 16 for i in resample_target_size]
            original_spacing_t = original_spacing[transpose_forward]
            new_spacing = (
                (np.array(resample_target_size).astype(float) / image_shape_t.astype(float))
                * np.array(original_spacing_t).astype(float)
            ).tolist()

        # Otherwise we need to calculate a new target shape, and we need to factor in that
        # the images will first be transposed and THEN resampled.
        # Find new shape based on the target spacing
        elif self.target_spacing is not None:
            target_spacing = np.array(self.target_spacing, dtype=float)
            original_spacing_t = original_spacing[transpose_forward]
            target_spacing_t = target_spacing[transpose_forward]
            resample_target_size = np.round((original_spacing_t / target_spacing_t).astype(float) * image_shape_t).astype(int)
            new_spacing = target_spacing_t.tolist()
        else:
            resample_target_size = image_shape_t
        return resample_target_size, final_target_size, new_spacing

    def resample_and_normalize_case(
        self,
        images: list,
        target_size,
        label: np.ndarray = None,
        norm_op=None,
    ):
        # Normalize and Transpose images to target view.
        # Transpose labels to target view.
        assert len(images) == len(norm_op) == len(self.intensities), (
            "number of images, "
            "normalization  operations and intensities does not match. \n"
            f"len(images) == {len(images)} \n"
            f"len(norm_op) == {len(norm_op)} \n"
            f"len(self.intensities) == {len(self.intensities)} \n"
        )

        for i in range(len(images)):
            image = images[i]
            assert image is not None
            images[i] = normalizer(image, scheme=norm_op[i], intensities=self.intensities[i])

        # Resample to target shape and spacing
        for i in range(len(images)):
            try:
                images[i] = resize(images[i], output_shape=target_size, order=3)
            except OverflowError:
                logging.error("Unexpected values in either shape or image for resize")
        if label is not None:
            try:
                label = resize(label, output_shape=target_size, order=0, anti_aliasing=False)
            except OverflowError:
                logging.error("Unexpected values in either shape or label for resize")
            return images, label
        return images

    def run_sanity_checks(self, images, label, subject_id, imagepaths):
        self.sanity_check_standard_formats(images, label, subject_id, imagepaths)

        if isinstance(images[0], nib.Nifti1Image):
            self.sanity_check_niftis(images, label, subject_id)

    def sanity_check_standard_formats(self, images, label, subject_id, imagepaths):
        assert len(images) > 0, f"found no images for {subject_id + '_'}, " f"attempted imagepaths: {imagepaths}"

        assert (
            len(images[0].shape) == self.plans["dataset_properties"]["data_dimensions"]
        ), f"image should be shape (x, y(, z)) but is {images[0].shape}"

        if label is not None:
            # make sure images and labels are correctly registered
            assert images[0].shape == label.shape, (
                f"Sizes do not match for {subject_id}" f"Image is: {images[0].shape} while the label is {label.shape}"
            )
        # Make sure all modalities are correctly registered
        if len(images) > 1:
            for image in images:
                assert images[0].shape == image.shape, (
                    f"Sizes do not match for {subject_id}" f"One is: {images[0].shape} while another is {image.shape}"
                )

    def sanity_check_niftis(self, images, label, subject_id):
        if label is not None:
            assert np.allclose(get_nib_spacing(images[0]), get_nib_spacing(label)), (
                f"Spacings do not match for {subject_id}"
                f"Image is: {get_nib_spacing(images[0])} while the label is {get_nib_spacing(label)}"
            )

            assert get_nib_orientation(images[0]) == get_nib_orientation(label), (
                f"Directions do not match for {subject_id}"
                f"Image is: {get_nib_orientation(images[0])} while the label is {get_nib_orientation(label)}"
            )
        if len(images) > 1:
            for image in images:
                assert np.allclose(get_nib_spacing(images[0]), get_nib_spacing(image)), (
                    f"Spacings do not match for {subject_id}"
                    f"One is: {get_nib_spacing(images[0])} while another is {get_nib_spacing(image)}"
                )

                assert get_nib_orientation(images[0]) == get_nib_orientation(image), (
                    f"Directions do not match for {subject_id}"
                    f"One is: {get_nib_orientation(images[0])} while another is {get_nib_orientation(image)}"
                )

    def verify_label_validity(self, label, subject_id):
        # Check if the ground truth only contains expected values
        expected_labels = np.array(self.plans["dataset_properties"]["classes"], dtype=np.float32)
        actual_labels = np.unique(label).astype(np.float32)
        assert np.all(np.isin(actual_labels, expected_labels)), (
            f"Unexpected labels found for {subject_id} \n" f"expected: {expected_labels} \n" f"found: {actual_labels}"
        )



from yucca.image_processing.transforms.YuccaTransform import YuccaTransform
import numpy as np


class BiasField(YuccaTransform):
    def __init__(self, data_key="image", p_per_sample=1):
        self.data_key = data_key
        self.p_per_sample = p_per_sample

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __biasField__(self, imageVolume):
        if len(imageVolume.shape) == 3:
            x, y, z = imageVolume.shape
            X, Y, Z = np.meshgrid(
                np.linspace(0, x, x, endpoint=False),
                np.linspace(0, y, y, endpoint=False),
                np.linspace(0, z, z, endpoint=False),
                indexing="ij",
            )
            x0 = np.random.randint(0, x)
            y0 = np.random.randint(0, y)
            z0 = np.random.randint(0, z)
            G = 1 - (np.power((X - x0), 2) / (x**2) + np.power((Y - y0), 2) / (y**2) + np.power((Z - z0), 2) / (z**2))
        else:
            x, y = imageVolume.shape
            X, Y = np.meshgrid(
                np.linspace(0, x, x, endpoint=False),
                np.linspace(0, y, y, endpoint=False),
                indexing="ij",
            )
            x0 = np.random.randint(0, x)
            y0 = np.random.randint(0, y)
            G = 1 - (np.power((X - x0), 2) / (x**2) + np.power((Y - y0), 2) / (y**2))
        return np.multiply(G, imageVolume)

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape. \nShould be (b, c, x, y, z) or (b, c, x, y) and is:\
                {data_dict[self.data_key].shape}"

        for b in range(data_dict[self.data_key].shape[0]):
            for c in range(data_dict[self.data_key][b].shape[0]):
                if np.random.uniform() < self.p_per_sample:
                    data_dict[self.data_key][b, c] = self.__biasField__(data_dict[self.data_key][b, c])
        return data_dict



from torch import nn
import numpy as np


class DeepSupervisionLoss(nn.Module):
    """
    Takes in a list of logits and a list of downsampled segmentations.
    Expects the first entry in each list to be the full-size/original segmentation, which will
    also be the one with the highest weight
    """

    def __init__(self, loss, weights=None):
        super().__init__()
        self.loss = loss
        self.weights = weights

    def forward(self, list_of_logits, list_of_segs):
        if self.weights is None:
            weights = np.array([1 / (2**i) for i in range(len(list_of_logits))])
            self.weights = weights / np.sum(weights)
        loss = self.loss(list_of_logits[0], list_of_segs[0]) * self.weights[0]
        for i in range(1, len(list_of_logits)):
            loss += self.loss(list_of_logits[i], list_of_segs[i]) * self.weights[i]
        return loss



import torch.nn as nn
from monai.networks.blocks import UnetrBasicBlock, UnetrUpBlock
from monai.networks.blocks.dynunet_block import UnetOutBlock
from yucca.network_architectures.networks.YuccaNet import YuccaNet
from yucca.network_architectures.blocks_and_layers.conv_blocks import UXNet_encoder
from yucca.network_architectures.blocks_and_layers.norm import (
    LayerNorm2d,
    LayerNorm3d,
)


class UXNet(YuccaNet):
    "https://github.com/MASILab/3DUX-Net/blob/main/networks/UXNet_3D/uxnet_encoder.py"

    def __init__(
        self,
        input_channels: int,
        num_classes: int = 1,
        conv_op=nn.Conv2d,
        weightInitializer=None,
        deep_supervision=False,
    ) -> None:
        """
        Args:
            in_channels: dimension of input channels.
            num_classes: dimension of output channels.
            conv_block: bool argument to determine if convolutional block is used.
            res_block: bool argument to determine if residual block is used.
            dropout_rate: faction of the input units to drop.


        """
        super(UXNet, self).__init__()

        if conv_op == nn.Conv2d:
            self.norm_op = LayerNorm2d
            self.spatial_dims = 2

        else:
            self.norm_op = LayerNorm3d
            self.spatial_dims = 3

        # Fixed parameters
        self.hidden_size = 768
        norm_name = "instance"
        res_block = True
        self.drop_path_rate = 0
        self.depths = [2, 2, 2, 2]
        self.feat_size = [48, 96, 192, 384]

        # Task specific
        self.num_classes = num_classes

        # Model parameters
        self.conv_op = conv_op
        self.norm_op_kwargs = {"eps": 1e-5}
        self.weightInitializer = weightInitializer
        self.deep_supervision = deep_supervision

        self.num_layers = 12

        self.out_indice = []
        for i in range(len(self.feat_size)):
            self.out_indice.append(i)

        self.uxnet = UXNet_encoder(
            in_channels=input_channels,
            depths=self.depths,
            dims=self.feat_size,
            conv_op=self.conv_op,
            norm_op=self.norm_op,
            drop_path_rate=self.drop_path_rate,
            layer_scale_init_value=1e-6,
            out_indices=self.out_indice,
        )

        self.encoder1 = UnetrBasicBlock(
            spatial_dims=self.spatial_dims,
            in_channels=input_channels,
            out_channels=self.feat_size[0],
            kernel_size=3,
            stride=1,
            norm_name=norm_name,
            res_block=res_block,
        )

        self.encoder2 = UnetrBasicBlock(
            spatial_dims=self.spatial_dims,
            in_channels=self.feat_size[0],
            out_channels=self.feat_size[1],
            kernel_size=3,
            stride=1,
            norm_name=norm_name,
            res_block=res_block,
        )

        self.encoder3 = UnetrBasicBlock(
            spatial_dims=self.spatial_dims,
            in_channels=self.feat_size[1],
            out_channels=self.feat_size[2],
            kernel_size=3,
            stride=1,
            norm_name=norm_name,
            res_block=res_block,
        )

        self.encoder4 = UnetrBasicBlock(
            spatial_dims=self.spatial_dims,
            in_channels=self.feat_size[2],
            out_channels=self.feat_size[3],
            kernel_size=3,
            stride=1,
            norm_name=norm_name,
            res_block=res_block,
        )

        self.encoder5 = UnetrBasicBlock(
            spatial_dims=self.spatial_dims,
            in_channels=self.feat_size[3],
            out_channels=self.hidden_size,
            kernel_size=3,
            stride=1,
            norm_name=norm_name,
            res_block=res_block,
        )

        self.ds_out_conv0 = self.conv_op(self.hidden_size, self.num_classes, kernel_size=1)

        self.decoder1 = UnetrUpBlock(
            spatial_dims=self.spatial_dims,
            in_channels=self.hidden_size,
            out_channels=self.feat_size[3],
            kernel_size=3,
            upsample_kernel_size=2,
            norm_name=norm_name,
            res_block=res_block,
        )

        self.ds_out_conv1 = self.conv_op(self.feat_size[3], self.num_classes, kernel_size=1)

        self.decoder2 = UnetrUpBlock(
            spatial_dims=self.spatial_dims,
            in_channels=self.feat_size[3],
            out_channels=self.feat_size[2],
            kernel_size=3,
            upsample_kernel_size=2,
            norm_name=norm_name,
            res_block=res_block,
        )

        self.ds_out_conv2 = self.conv_op(self.feat_size[2], self.num_classes, kernel_size=1)

        self.decoder3 = UnetrUpBlock(
            spatial_dims=self.spatial_dims,
            in_channels=self.feat_size[2],
            out_channels=self.feat_size[1],
            kernel_size=3,
            upsample_kernel_size=2,
            norm_name=norm_name,
            res_block=res_block,
        )

        self.ds_out_conv3 = self.conv_op(self.feat_size[1], self.num_classes, kernel_size=1)

        self.decoder4 = UnetrUpBlock(
            spatial_dims=self.spatial_dims,
            in_channels=self.feat_size[1],
            out_channels=self.feat_size[0],
            kernel_size=3,
            upsample_kernel_size=2,
            norm_name=norm_name,
            res_block=res_block,
        )

        self.decoder5 = UnetrBasicBlock(
            spatial_dims=self.spatial_dims,
            in_channels=self.feat_size[0],
            out_channels=self.feat_size[0],
            kernel_size=3,
            stride=1,
            norm_name=norm_name,
            res_block=res_block,
        )

        self.out = UnetOutBlock(spatial_dims=self.spatial_dims, in_channels=self.feat_size[0], out_channels=self.num_classes)  # type: ignore

        if self.weightInitializer is not None:
            print("initializing weights")
            self.apply(self.weightInitializer)

    def forward(self, x_in):
        outs = self.uxnet(x_in)

        enc1 = self.encoder1(x_in)
        x2 = outs[0]
        enc2 = self.encoder2(x2)
        x3 = outs[1]
        enc3 = self.encoder3(x3)
        x4 = outs[2]
        enc4 = self.encoder4(x4)
        enc_hidden = self.encoder5(outs[3])

        dec1 = self.decoder1(enc_hidden, enc4)
        dec2 = self.decoder2(dec1, enc3)
        dec3 = self.decoder3(dec2, enc2)
        dec4 = self.decoder4(dec3, enc1)
        dec5 = self.decoder5(dec4)

        if self.deep_supervision:
            ds0 = self.ds_out_conv0(enc_hidden)
            ds1 = self.ds_out_conv1(dec1)
            ds2 = self.ds_out_conv2(dec2)
            ds3 = self.ds_out_conv3(dec3)
            ds4 = self.out(dec5)
            return [ds4, ds3, ds2, ds1, ds0]

        return self.out(dec5)



import yaml
import os
import nibabel as nib
import numpy as np
from PIL import Image


def load_yaml(file: str):
    with open(file, "r") as f:
        a = yaml.load(f, Loader=yaml.BaseLoader)
    return a


def read_file_to_nifti_or_np(imagepath, dtype=np.float32):
    ext = imagepath.split(os.extsep, 1)[1]
    if ext in ["nii", "nii.gz"]:
        return nib.load(imagepath)
    elif ext in ["png", "jpg", "jpeg"]:
        return np.array(Image.open(imagepath).convert("L"), dtype=dtype)
    elif ext in ["csv", "txt"]:
        return np.atleast_1d(np.genfromtxt(imagepath, delimiter=",", dtype=dtype))
    else:
        raise TypeError(f"File type invalid. Found extension: {ext} and expected one in [nii, nii.gz, png, csv, txt]")



import torch
import numpy as np
import yucca
import wandb
from sklearn.model_selection import KFold
from torch import optim
from torch.cuda.amp import GradScaler
from batchgenerators.utilities.file_and_folder_operations import (
    join,
    subfiles,
    save_pickle,
)
from yucca.paths import yucca_preprocessed_data
from yucca.training.trainers.base_manager import base_manager
from yucca.evaluation.confusion_matrix import (
    torch_confusion_matrix_from_logits,
    torch_get_tp_fp_tn_fn,
)
from yucca.evaluation.metrics import dice_per_label
from yucca.loss_and_optim.loss_functions.nnUNet_losses import DiceCE
from yucca.utils.files_and_folders import recursive_find_python_class
from yucca.utils.kwargs import filter_kwargs


class YuccaManager(base_manager):
    """
    This is the barebone functional trainer of the pipeline.

    When training is started it will call self.initialize(). This carries out the following steps:
        - Loads the plans in the folder of the preprocessed training data
        - Sets the required outpaths for saving model weights, training parameters and logs
        - Sets the # epochs and # batches pr. epoch
            - By default 1000 epochs of 250/50 train/val batches pr. epoch
        - Sets the batch size and patch size
          - Patch Size is automatically set based on the dataset.
        - Defines the Data Augmentation scheme
        - Initializes the network
            - Input and output channels are derived from the # modalities and
            # classes stored in the plans file of the dataset.
        - Loads the data
            - If no splits are provided it will automatically split the data.
            - This initializes the data loaders and feeds them to the YuccaAugmenter
        - Initializes the loss function, optimizer, learning rate, scheduler and scaler.
        - Saves all these parameters in the defined outpath
        - Initializes weights & biases
    Then it starts training
        - Using mixed precision
        - Gradients are scaled and clipped.
        - Augmentations are applied using the CPU while the forward - backward pass is handled
        by the GPU to avoid IO bottlenecks
        - On epoch start the start time is recorded and lists for losses etc. are created
        - On epoch finish the end time is recorded and validation results are evaluated and uploaded
        to both WANDB and the local training log. Finally the LR is updated according to any
        scheduler.
        - Model weights are stored every Xth (default 300) epoch AND whenever the model improves its historically
        best Dice on the validation set.
    After the last epoch the final model weights are saved and training is concluded.
    """

    def __init__(
        self,
        model,
        model_dimensions: str,
        task: str,
        folds: str | int,
        planner: str,
        starting_lr: float = None,
        loss_fn: str = None,
        momentum: float = None,
        continue_training: bool = False,
        checkpoint: str = None,
        finetune: bool = False,
        fast_training: bool = False,
    ):
        super().__init__()

        # Trainer specific parameters
        self._DEFAULT_STARTING_LR = 1e-3
        self._DEFAULT_LOSS = DiceCE
        self._DEFAULT_MOMENTUM = 0.9
        self.batch_size_3D = 2
        self.batch_size_2D = 64
        self.train_batches_per_epoch = 250
        self.val_batches_per_epoch = 50
        self.name = self.__class__.__name__
        self.optim = optim.SGD
        self.lr_scheduler = optim.lr_scheduler.CosineAnnealingLR
        self.final_epoch = 1000
        self.p_force_foreground = 0.33
        self.fast_final_epoch = 500
        self.fast_train_batches_per_epoch = 100
        self.fast_val_batches_per_epoch = 20
        self.max_vram = 12

        # REQUIRED:
        # These must be defined in run_training
        self.model_name = model
        self.model_dimensions = model_dimensions
        self.task = task
        self.folds = int(folds)
        self.plan_id = planner
        self.continue_training = continue_training
        self.checkpoint = checkpoint
        self.fast_training = fast_training
        self.finetune = finetune

        # OPTIONALS:
        # These can be defined in run_training
        self.starting_lr = starting_lr
        self.loss_fn = loss_fn
        self.momentum = momentum

        # The default is for the basic UNet.
        self.patch_size_3D = (96, 96, 96)
        self.patch_size_2D = (160, 160)
        if self.model_name == "MultiResUNet":
            self.patch_size_3D = (96, 96, 80)
            self.patch_size_2D = (144, 144)

        # kwargs that we set when params such as LR has been settled (do we use default or is it
        # specified?)
        self.loss_fn_kwargs = {}
        self.optim_kwargs = {}
        self.lr_scheduler_kwargs = {}

    def comprehensive_eval(self, pred, seg):
        if not self.epoch_eval_dict:
            self.epoch_eval_dict = {
                stat: []
                for stat in [
                    "Dice           :",
                    "True Positives :",
                    "False Positives:",
                    "False Negatives:",
                ]
            }

        confusion_matrix = torch_confusion_matrix_from_logits(pred, seg)
        tp, fp, _, fn = torch_get_tp_fp_tn_fn(confusion_matrix, ignore_label=0)
        self.epoch_eval_dict["Dice           :"].append(dice_per_label(tp, fp, _, fn))
        self.epoch_eval_dict["True Positives :"].append(tp)
        self.epoch_eval_dict["False Positives:"].append(fp)
        self.epoch_eval_dict["False Negatives:"].append(fn)

    def initialize(self):
        if not self.is_initialized:
            # Then we load the plans and set modalities and classes etc.
            self.load_plans_from_path(
                join(
                    yucca_preprocessed_data,
                    self.task,
                    self.plan_id,
                    self.plan_id + "_plans.json",
                )
            )

            # First we set the path based on the supplied parameters
            self.set_outpath()

            # Now we can log the plan file, after the outpath has been created
            self.log(f'{"plan file:":20}', self.plans_path, time=False)

            # Now we set up parameters for the data and the data augmentations
            self.set_train_length()
            self.set_batch_and_patch_sizes()
            self.setup_DA()

            # Load self.network based on the
            # dimensions (2D or 3D) and classes (number of output channels)
            self.initialize_network()
            self.log(
                f'{"network:":20}',
                self.model_dimensions,
                self.network.__class__.__name__,
                time=False,
            )

            # Load the data (and if necessary split it)
            # Before loading data, we set random seeds for reproducibility
            self.load_data()

            # Setup optimizer, learning rate, momentum etc.
            self.initialize_loss_optim_lr()

            # Before saving parameters and moving on to training
            # check if we are using pretrained weights/continuing training
            if self.continue_training:
                if self.checkpoint:
                    self.load_checkpoint(self.checkpoint, train=True)
                else:
                    latest = subfiles(self.outpath, suffix="latest.model", join=False)[0]
                    assert latest, "Can not continue training. No latest checkpoint found."

                    self.load_checkpoint(join(self.outpath, latest), train=True)

            # And finally save a json with arguments and parameters
            self.save_parameter_json()
            self.initialize_wandb()

            self.log("initialization done, printing network", "\n", time=False)
            self.log(self.network, time=False)

            self.is_initialized = True
        else:
            print(
                "Network is already initialized. \
                  Calling initialize repeatedly should be avoided."
            )

    def initialize_loss_optim_lr(self):
        # Define the gradscaler
        self.grad_scaler = GradScaler(enabled=True)

        # Defining the loss
        if not self.loss_fn:
            self.loss_fn = self._DEFAULT_LOSS

        elif self.loss_fn:
            self.loss_fn = recursive_find_python_class(
                folder=[join(yucca.__path__[0], "training", "loss_functions")],
                class_name=self.loss_fn,
                current_module="yucca.training.loss_functions",
            )

        self.loss_fn_kwargs = {
            "soft_dice_kwargs": {"apply_softmax": True},  # DCE
            "hierarchical_kwargs": {"rootdict": self.plans["dataset_properties"]["label_hierarchy"]},  # Hierarchical Loss
        }
        self.loss_fn_kwargs = filter_kwargs(self.loss_fn, self.loss_fn_kwargs)
        self.loss_fn = self.loss_fn(**self.loss_fn_kwargs)
        assert isinstance(self.loss_fn, torch.nn.Module), (
            "Loss is not a torch.nn.Module. " "Make sure the correct loss was found (check spelling)"
        )

        self.log(f'{"loss function":20}', self.loss_fn.__class__.__name__, time=False)

        # Defining the learning rate
        if not self.starting_lr:
            self.starting_lr = self._DEFAULT_STARTING_LR
        if self.finetune:
            self.starting_lr /= 10
        self.log(f'{"learning rate:":20}', self.starting_lr, time=False)

        # Defining the momentum
        if not self.momentum:
            self.momentum = self._DEFAULT_MOMENTUM
        self.log(f'{"momentum":20}', self.momentum, time=False)

        # And constructing the optimizer
        # Set kwargs for all optimizers and then filter relevant ones based on optimizer class

        self.optim_kwargs = {
            "lr": float(self.starting_lr),  # all
            "momentum": float(self.momentum),  # SGD
            "eps": 1e-4,
            "weight_decay": 3e-5,
        }
        self.optim_kwargs = filter_kwargs(self.optim, self.optim_kwargs)
        self.optim = self.optim(self.network.parameters(), **self.optim_kwargs)
        self.log(f'{"optimizer":20}', self.optim.__class__.__name__, time=False)

        # Set kwargs for all schedulers and then filter relevant ones based on scheduler class
        self.lr_scheduler_kwargs = {
            "T_max": self.final_epoch,
            "eta_min": 1e-9,  # Cosine Annealing
        }
        self.lr_scheduler_kwargs = filter_kwargs(self.lr_scheduler, self.lr_scheduler_kwargs)
        self.lr_scheduler = self.lr_scheduler(self.optim, **self.lr_scheduler_kwargs)
        self.log(f'{"LR scheduler":20}', self.lr_scheduler.__class__.__name__, time=False)

    def run_training(self):
        self.initialize()

        _ = self.tr_gen.next()
        _ = self.val_gen.next()

        if torch.cuda.is_available():
            torch.cuda.empty_cache()
        else:
            self.log(
                "CUDA NOT AVAILABLE. YOU SHOULD REALLY NOT TRAIN WITHOUT CUDA!",
                time=False,
            )

        wandb.watch(self.network)

        while self.current_epoch < self.final_epoch:
            self.epoch_start()

            self.network.train()
            for _ in range(self.train_batches_per_epoch):
                batch_loss = self.run_batch(next(self.tr_gen))
                self.epoch_tr_loss.append(batch_loss)

            self.network.eval()
            with torch.no_grad():
                for _ in range(self.val_batches_per_epoch):
                    batch_loss = self.run_batch(next(self.val_gen), train=False, comprehensive_eval=True)
                    self.epoch_val_loss.append(batch_loss)

            self.tr_losses.append(np.mean(self.epoch_tr_loss))
            self.val_losses.append(np.mean(self.epoch_val_loss))

            if self.lr_scheduler:
                self.lr_scheduler.step()

            self.epoch_finish()

        if self.current_epoch == self.final_epoch:
            self.save_checkpoint("checkpoint_final.model")
            wandb.finish()

    def setup_DA(self):
        super().setup_DA()
        self.AdditiveNoise_p_per_sample = 0.2
        self.BiasField_p_per_sample = 0.33
        self.Blurring_p_per_sample = 0.2
        self.ElasticDeform_p_per_sample = 0.33
        self.Gamma_p_per_sample = 0.2
        self.GibbsRinging_p_per_sample = 0.2
        self.Mirror_p_per_sample = 0.2
        self.MotionGhosting_p_per_sample = 0.2
        self.MultiplicativeNoise_p_per_sample = 0.2
        self.Rotation_p_per_sample = 0.33
        self.Scale_p_per_sample = 0.33
        self.SimulateLowres_p_per_sample = 0.2

    def split_data(self):
        splits = []

        files = subfiles(self.folder_with_preprocessed_data, join=False, suffix=".npy")
        if not files:
            files = subfiles(self.folder_with_preprocessed_data, join=False, suffix=".npz")
            if files:
                self.log(
                    "Only found compressed (.npz) files. This might increase runtime.",
                    time=False,
                )

        assert files, f"Couldn't find any .npy or .npz files in {self.folder_with_preprocessed_data}"

        files = np.array(files)
        # We set this seed manually as multiple trainers might use this split,
        # And we may not know which individual seed dictated the data splits
        # Therefore for reproducibility this is fixed.

        kf = KFold(n_splits=5, shuffle=True, random_state=52189)
        for train, val in kf.split(files):
            splits.append({"train": list(files[train]), "val": list(files[val])})

        save_pickle(splits, self.splits_file)



no_aug = {
    "additive_noise_p_per_sample": 0.0,
    "biasfield_p_per_sample": 0.0,
    "blurring_p_per_sample": 0.0,
    "blurring_p_per_channel": 0.0,
    "elastic_deform_p_per_sample": 0.0,
    "gamma_p_per_sample": 0.0,
    "gamma_p_invert_image": 0.0,
    "gibbs_ringing_p_per_sample": 0.0,
    "mirror_p_per_sample": 0.0,
    "mirror_p_per_axis": 0.0,
    "motion_ghosting_p_per_sample": 0.0,
    "multiplicative_noise_p_per_sample": 0.0,
    "rotation_p_per_sample": 0.0,
    "rotation_p_per_axis": 0.0,
    "scale_p_per_sample": 0.0,
    "simulate_lowres_p_per_sample": 0.0,
    "simulate_lowres_p_per_channel": 0.0,
    "simulate_lowres_p_per_axis": 0.0,
}

all_always = {
    "random_crop": True,
    "mask_image_for_reconstruction": True,
    "skip_label": False,
    "label_dtype": int,
    "copy_image_to_label": False,
    "additive_noise_p_per_sample": 1.0,
    "biasfield_p_per_sample": 1.0,
    "blurring_p_per_sample": 1.0,
    "blurring_p_per_channel": 1.0,
    "elastic_deform_p_per_sample": 1.0,
    "gamma_p_per_sample": 1.0,
    "gamma_p_invert_image": 1.0,
    "gibbs_ringing_p_per_sample": 1.0,
    "mirror_p_per_sample": 1.0,
    "mirror_p_per_axis": 1.0,
    "motion_ghosting_p_per_sample": 1.0,
    "multiplicative_noise_p_per_sample": 1.0,
    "rotation_p_per_sample": 1.0,
    "rotation_p_per_axis": 1.0,
    "scale_p_per_sample": 1.0,
    "simulate_lowres_p_per_sample": 1.0,
    "simulate_lowres_p_per_channel": 1.0,
    "simulate_lowres_p_per_axis": 1.0,
}


basic = no_aug.copy()
basic["mirror_p_per_sample"] = 0.2
basic["mirror_p_per_axis"] = 0.66
basic["rotation_p_per_sample"] = 0.2
basic["rotation_p_per_axis"] = 0.66
basic["scale_p_per_sample"] = 0.2



import nibabel as nib
import numpy as np
from lightning.pytorch.callbacks import BasePredictionWriter
from yucca.utils.softmax import softmax
from yucca.utils.nib_utils import reorient_nib_image
from PIL import Image
from batchgenerators.utilities.file_and_folder_operations import (
    join,
    subfiles,
    maybe_mkdir_p,
)


def save_nifti_from_numpy(pred, outpath, properties, compression=9):
    # slight hacky, but it is what it is
    nib.openers.Opener.default_compresslevel = compression
    pred = nib.Nifti1Image(
        pred, properties["nifti_metadata"]["affine"], header=properties["nifti_metadata"]["header"], dtype=np.uint8
    )
    if properties["nifti_metadata"]["reoriented"]:
        pred = reorient_nib_image(
            pred, properties["nifti_metadata"]["final_direction"], properties["nifti_metadata"]["original_orientation"]
        )
    nib.save(
        pred,
        outpath + ".nii.gz",
    )
    del pred


def save_png_from_numpy(pred, outpath):
    pred = Image.fromarray(pred)
    pred.save(outpath)
    del pred


def save_txt_from_numpy(pred, outpath):
    np.savetxt(outpath + ".txt", np.atleast_1d(pred), fmt="%i", delimiter=",")
    del pred


def save_prediction_from_logits(logits, outpath, properties, save_softmax=False, compression=9):
    if save_softmax:
        softmax_result = softmax(logits)[0].astype(np.float32)
        np.savez_compressed(outpath + ".npz", data=softmax_result, properties=properties)
    if logits.shape[1] > 1:
        logits = np.argmax(logits, 1)
    pred = np.squeeze(logits)
    if properties.get("save_format") == "png":
        save_png_from_numpy(pred, outpath)
    if properties.get("save_format") == "txt":
        save_txt_from_numpy(pred, outpath)
    else:
        save_nifti_from_numpy(pred, outpath, properties, compression=compression)


def merge_softmax_from_folders(folders: list, outpath, method="sum"):
    maybe_mkdir_p(outpath)
    cases = subfiles(folders[0], suffix=".npz", join=False)
    for folder in folders:
        assert cases == subfiles(folder, suffix=".npz", join=False), (
            f"Found unexpected cases. "
            f"The following two folders do not contain the same cases: \n"
            f"{folders[0]} \n"
            f"{folder}"
        )

    for case in cases:
        files_for_case = [np.load(join(folder, case), allow_pickle=True) for folder in folders]
        properties_for_case = files_for_case[0]["properties"]
        files_for_case = [file["data"].astype(np.float32) for file in files_for_case]

        if method == "sum":
            files_for_case = np.sum(files_for_case, axis=0)

        files_for_case = np.argmax(files_for_case, 0)
        save_nifti_from_numpy(
            files_for_case,
            join(outpath, case[:-4]),
            properties=properties_for_case.item(),
        )

    del files_for_case, properties_for_case


class WritePredictionFromLogits(BasePredictionWriter):
    def __init__(self, output_dir, save_softmax, write_interval):
        super().__init__(write_interval)
        self.output_dir = output_dir
        self.save_softmax = save_softmax

    def write_on_batch_end(self, _trainer, _pl_module, data_dict, _batch_indices, _batch, _batch_idx, _dataloader_idx):
        # this will create N (num processes) files in `output_dir` each containing
        # the predictions of it's respective rank
        logits, properties, case_id = (
            data_dict["logits"],
            data_dict["properties"],
            data_dict["case_id"],
        )
        save_prediction_from_logits(
            logits,
            join(self.output_dir, case_id),
            properties=properties,
            save_softmax=self.save_softmax,
        )
        del data_dict



import numpy as np
from yucca.image_processing.transforms.YuccaTransform import YuccaTransform
from typing import Literal, Union


class CropPad(YuccaTransform):
    def __init__(
        self,
        data_key="image",
        label_key="label",
        pad_value: Union[Literal["min", "zero", "edge"], int, float] = "min",
        patch_size: tuple | list = None,
        p_oversample_foreground=0.0,
    ):
        self.data_key = data_key
        self.label_key = label_key
        self.pad_value = pad_value
        self.patch_size = patch_size
        self.p_oversample_foreground = p_oversample_foreground

    @staticmethod
    def get_params(data, pad_value, target_shape):
        if pad_value == "min":
            pad_kwargs = {"constant_values": data.min(), "mode": "constant"}
        elif pad_value == "zero":
            pad_kwargs = {"constant_values": np.zeros(1, dtype=data.dtype), "mode": "constant"}
        elif isinstance(pad_value, int) or isinstance(pad_value, float):
            pad_kwargs = {"constant_values": pad_value, "mode": "constant"}
        elif pad_value == "edge":
            pad_kwargs = {"mode": "edge"}
        else:
            print("Unrecognized pad value detected.")
        input_shape = data.shape
        target_image_shape = (input_shape[0], *target_shape)
        target_label_shape = (1, *target_shape)
        return input_shape, target_image_shape, target_label_shape, pad_kwargs

    def __croppad__(
        self,
        data_dict: np.ndarray,
        image_properties: dict,
        input_shape: np.ndarray,
        target_image_shape: list | tuple,
        target_label_shape: list | tuple,
        **pad_kwargs,
    ):
        image = data_dict[self.data_key]
        if data_dict.get(self.label_key) is not None:
            label = data_dict[self.label_key]
        else:
            label = None

        if len(self.patch_size) == 3:
            image, label = self.generate_3D_case_from_3D(
                image, image_properties, label, target_image_shape, target_label_shape, **pad_kwargs
            )
        elif len(self.patch_size) == 2 and len(input_shape) == 4:
            image, label = self.generate_2D_case_from_3D(
                image, image_properties, label, target_image_shape, target_label_shape, **pad_kwargs
            )
        elif len(self.patch_size) == 2 and len(input_shape) == 3:
            image, label = self.generate_2D_case_from_2D(
                image, image_properties, label, target_image_shape, target_label_shape, **pad_kwargs
            )
        data_dict[self.data_key] = image
        if label is not None:
            data_dict[self.label_key] = label
        return data_dict

    def generate_3D_case_from_3D(self, image, image_properties, label, target_image_shape, target_label_shape, **pad_kwargs):
        image_out = np.zeros(target_image_shape)
        label_out = np.zeros(target_label_shape)

        # First we pad to ensure min size is met
        to_pad = []
        for d in range(3):
            if image.shape[d + 1] < self.patch_size[d]:
                to_pad += [(self.patch_size[d] - image.shape[d + 1])]
            else:
                to_pad += [0]

        pad_lb_x = to_pad[0] // 2
        pad_ub_x = to_pad[0] // 2 + to_pad[0] % 2
        pad_lb_y = to_pad[1] // 2
        pad_ub_y = to_pad[1] // 2 + to_pad[1] % 2
        pad_lb_z = to_pad[2] // 2
        pad_ub_z = to_pad[2] // 2 + to_pad[2] % 2

        # This is where we should implement any patch selection biases.
        # The final patch excted after augmentation will always be the center of this patch
        # to avoid interpolation artifacts near the borders
        crop_start_idx = []
        if len(image_properties["foreground_locations"]) == 0 or np.random.uniform() >= self.p_oversample_foreground:
            for d in range(3):
                if image.shape[d + 1] < self.patch_size[d]:
                    crop_start_idx += [0]
                else:
                    crop_start_idx += [np.random.randint(image.shape[d + 1] - self.patch_size[d] + 1)]
        else:
            locidx = np.random.choice(len(image_properties["foreground_locations"]))
            location = image_properties["foreground_locations"][locidx]
            for d in range(3):
                if image.shape[d + 1] < self.patch_size[d]:
                    crop_start_idx += [0]
                else:
                    crop_start_idx += [
                        np.random.randint(
                            max(0, location[d] - self.patch_size[d]),
                            min(location[d], image.shape[d + 1] - self.patch_size[d]) + 1,
                        )
                    ]

        image_out[
            :,
            :,
            :,
            :,
        ] = np.pad(
            image[
                :,
                crop_start_idx[0] : crop_start_idx[0] + self.patch_size[0],
                crop_start_idx[1] : crop_start_idx[1] + self.patch_size[1],
                crop_start_idx[2] : crop_start_idx[2] + self.patch_size[2],
            ],
            ((0, 0), (pad_lb_x, pad_ub_x), (pad_lb_y, pad_ub_y), (pad_lb_z, pad_ub_z)),
            **pad_kwargs,
        )
        if label is None:
            return image_out, None
        label_out[
            :,
            :,
            :,
            :,
        ] = np.pad(
            label[
                :,
                crop_start_idx[0] : crop_start_idx[0] + self.patch_size[0],
                crop_start_idx[1] : crop_start_idx[1] + self.patch_size[1],
                crop_start_idx[2] : crop_start_idx[2] + self.patch_size[2],
            ],
            ((0, 0), (pad_lb_x, pad_ub_x), (pad_lb_y, pad_ub_y), (pad_lb_z, pad_ub_z)),
        )
        return image_out, label_out

    def generate_2D_case_from_3D(self, image, image_properties, label, target_image_shape, target_label_shape, **pad_kwargs):
        """
        The possible input for this can be 2D or 3D data.
        For 2D we want to pad or crop as necessary.
        For 3D we want to first select a slice from the first dimension, i.e. volume[idx, :, :],
        then pad or crop as necessary.
        """
        image_out = np.zeros(target_image_shape)
        label_out = np.zeros(target_label_shape)

        # First we pad to ensure min size is met
        to_pad = []
        for d in range(2):
            if image.shape[d + 2] < self.patch_size[d]:
                to_pad += [(self.patch_size[d] - image.shape[d + 2])]
            else:
                to_pad += [0]

        pad_lb_y = to_pad[0] // 2
        pad_ub_y = to_pad[0] // 2 + to_pad[0] % 2
        pad_lb_z = to_pad[1] // 2
        pad_ub_z = to_pad[1] // 2 + to_pad[1] % 2

        # This is where we should implement any patch selection biases.
        # The final patch extracted after augmentation will always be the center of this patch
        # as this is where augmentation-induced interpolation artefacts are least likely
        crop_start_idx = []
        if len(image_properties["foreground_locations"]) == 0 or np.random.uniform() >= self.p_oversample_foreground:
            x_idx = np.random.randint(image.shape[1])
            for d in range(2):
                if image.shape[d + 2] < self.patch_size[d]:
                    crop_start_idx += [0]
                else:
                    crop_start_idx += [np.random.randint(image.shape[d + 2] - self.patch_size[d] + 1)]
        else:
            locidx = np.random.choice(len(image_properties["foreground_locations"]))
            location = image_properties["foreground_locations"][locidx]
            x_idx = location[0]
            for d in range(2):
                if image.shape[d + 2] < self.patch_size[d]:
                    crop_start_idx += [0]
                else:
                    crop_start_idx += [
                        np.random.randint(
                            max(0, location[d + 1] - self.patch_size[d]),
                            min(location[d + 1], image.shape[d + 2] - self.patch_size[d]) + 1,
                        )
                    ]

        image_out[:, :, :] = np.pad(
            image[
                :,
                x_idx,
                crop_start_idx[0] : crop_start_idx[0] + self.patch_size[0],
                crop_start_idx[1] : crop_start_idx[1] + self.patch_size[1],
            ],
            ((0, 0), (pad_lb_y, pad_ub_y), (pad_lb_z, pad_ub_z)),
            **pad_kwargs,
        )

        if label is None:
            return image_out, None

        label_out[:, :, :] = np.pad(
            label[
                :,
                x_idx,
                crop_start_idx[0] : crop_start_idx[0] + self.patch_size[0],
                crop_start_idx[1] : crop_start_idx[1] + self.patch_size[1],
            ],
            ((0, 0), (pad_lb_y, pad_ub_y), (pad_lb_z, pad_ub_z)),
        )

        return image_out, label_out

    def generate_2D_case_from_2D(self, image, image_properties, label, target_image_shape, target_label_shape, **pad_kwargs):
        """
        The possible input for this can be 2D or 3D data.
        For 2D we want to pad or crop as necessary.
        For 3D we want to first select a slice from the first dimension, i.e. volume[idx, :, :],
        then pad or crop as necessary.
        """
        image_out = np.zeros(target_image_shape)
        label_out = np.zeros(target_label_shape)

        # First we pad to ensure min size is met
        to_pad = []
        for d in range(2):
            if image.shape[d + 1] < self.patch_size[d]:
                to_pad += [(self.patch_size[d] - image.shape[d + 1])]
            else:
                to_pad += [0]

        pad_lb_x = to_pad[0] // 2
        pad_ub_x = to_pad[0] // 2 + to_pad[0] % 2
        pad_lb_y = to_pad[1] // 2
        pad_ub_y = to_pad[1] // 2 + to_pad[1] % 2

        # This is where we should implement any patch selection biases.
        # The final patch extracted after augmentation will always be the center of this patch
        # as this is where artefacts are least present
        crop_start_idx = []
        if len(image_properties["foreground_locations"]) == 0 or np.random.uniform() >= self.p_oversample_foreground:
            for d in range(2):
                if image.shape[d + 1] < self.patch_size[d]:
                    crop_start_idx += [0]
                else:
                    crop_start_idx += [np.random.randint(image.shape[d + 1] - self.patch_size[d] + 1)]
        else:
            locidx = np.random.choice(len(image_properties["foreground_locations"]))
            location = image_properties["foreground_locations"][locidx]
            for d in range(2):
                if image.shape[d + 1] < self.patch_size[d]:
                    crop_start_idx += [0]
                else:
                    crop_start_idx += [
                        np.random.randint(
                            max(0, location[d] - self.patch_size[d]),
                            min(location[d], image.shape[d + 1] - self.patch_size[d]) + 1,
                        )
                    ]

        image_out[:, :, :] = np.pad(
            image[
                :,
                crop_start_idx[0] : crop_start_idx[0] + self.patch_size[0],
                crop_start_idx[1] : crop_start_idx[1] + self.patch_size[1],
            ],
            ((0, 0), (pad_lb_x, pad_ub_x), (pad_lb_y, pad_ub_y)),
            **pad_kwargs,
        )

        if label is None:  # Reconstruction/inpainting
            return image_out, None

        if len(label.shape) == 1:  # Classification
            return image_out, label

        label_out[:, :, :] = np.pad(
            label[
                :,
                crop_start_idx[0] : crop_start_idx[0] + self.patch_size[0],
                crop_start_idx[1] : crop_start_idx[1] + self.patch_size[1],
            ],
            ((0, 0), (pad_lb_x, pad_ub_x), (pad_lb_y, pad_ub_y)),
        )

        return image_out, label_out

    def __call__(self, packed_data_dict=None, image_properties=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict

        input_shape, target_image_shape, target_label_shape, pad_kwargs = self.get_params(
            data=data_dict[self.data_key], pad_value=self.pad_value, target_shape=self.patch_size
        )

        data_dict = self.__croppad__(
            data_dict=data_dict,
            image_properties=image_properties,
            input_shape=input_shape,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )
        return data_dict



from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.task_conversion.utils import generate_dataset_json
import shutil
from yucca.paths import yucca_raw_data


def convert(path: str, subdir: str = "amos22"):
    # INPUT DATA

    path = f"{path}/{subdir}"
    file_suffix = ".nii.gz"

    # Train/Test Splits
    labels_dir_tr = join(path, "labelsTr")
    images_dir_tr = join(path, "imagesTr")
    training_samples = subfiles(labels_dir_tr, join=False, suffix=file_suffix)

    labels_dir_ts = join(path, "labelsVa")
    images_dir_ts = join(path, "imagesVa")
    test_samples = subfiles(labels_dir_ts, join=False, suffix=file_suffix)

    ###OUTPUT DATA
    # Target names
    task_name = "Task010_AMOS22"
    prefix = "AMOS22"

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    ###Populate Target Directory###
    # This is likely also the place to apply any re-orientation, resampling and/or label correction.
    for sTr in training_samples:
        serial_number = sTr[: -len(file_suffix)]
        shutil.copy2(join(images_dir_tr, sTr), f"{target_imagesTr}/{prefix}_{serial_number}_000.nii.gz")
        shutil.copy2(join(labels_dir_tr, sTr), f"{target_labelsTr}/{prefix}_{serial_number}.nii.gz")

    for sTs in test_samples:
        serial_number = sTs[: -len(file_suffix)]
        shutil.copy2(join(images_dir_ts, sTs), f"{target_imagesTs}/{prefix}_{serial_number}_000.nii.gz")
        shutil.copy2(join(labels_dir_ts, sTs), f"{target_labelsTs}/{prefix}_{serial_number}.nii.gz")

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        ("CT",),
        labels={
            "0": "background",
            "1": "spleen",
            "2": "right kidney",
            "3": "left kidney",
            "4": "gall bladder",
            "5": "esophagus",
            "6": "liver",
            "7": "stomach",
            "8": "arota",
            "9": "postcava",
            "10": "pancreas",
            "11": "right adrenal gland",
            "12": "left adrenal gland",
            "13": "duodenum",
            "14": "bladder",
            "15": "prostate/uterus",
        },
        dataset_name=task_name,
        license="hands off!",
        dataset_description="AMOS22",
        dataset_reference="",
    )



from yucca.image_processing.transforms.YuccaTransform import YuccaTransform
import numpy as np


class Mirror(YuccaTransform):
    def __init__(
        self,
        data_key="image",
        label_key="label",
        p_per_sample=1,
        axes=(0, 1, 2),
        p_mirror_per_axis=0.33,
        skip_label=False,
    ):
        self.data_key = data_key
        self.label_key = label_key
        self.p_per_sample = p_per_sample
        self.p_mirror_per_axis = p_mirror_per_axis
        self.axes = axes
        self.skip_label = skip_label

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __mirror__(self, data_dict, axes):
        image = data_dict[self.data_key]
        label = data_dict.get(self.label_key)
        # Input will be [b, c, x, y, z] or [b, c, x, y]
        for b in range(image.shape[0]):
            if np.random.uniform() < self.p_per_sample:
                if 0 in axes and np.random.uniform() < self.p_mirror_per_axis:
                    image[b, :, :] = image[b, :, ::-1]
                    if label is not None and not self.skip_label:
                        label[b, :, :] = label[b, :, ::-1]
                if 1 in axes and np.random.uniform() < self.p_mirror_per_axis:
                    image[b, :, :, :] = image[b, :, :, ::-1]
                    if label is not None and not self.skip_label:
                        label[b, :, :, :] = label[b, :, :, ::-1]
                if 2 in axes and np.random.uniform() < self.p_mirror_per_axis:
                    image[b, :, :, :, :] = image[b, :, :, :, ::-1]
                    if label is not None and not self.skip_label:
                        label[b, :, :, :, :] = label[b, :, :, :, ::-1]
        data_dict[self.data_key] = image
        if label is not None and not self.skip_label:
            data_dict[self.label_key] = label
        return data_dict

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
            \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.data_key].shape}"

        data_dict = self.__mirror__(
            data_dict,
            self.axes,
        )
        return data_dict



from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.task_conversion.utils import generate_dataset_json
import shutil
import gzip
from yucca.paths import yucca_raw_data
from multiprocessing import Pool


def convert_case(sample, input_image_dir, input_label_dir, target_image_dir, target_label_dir, prefix, suffix):
    serial_number = sample[: -len(suffix)]
    image_file = open(join(input_image_dir, sample), "rb")
    label = open(join(input_label_dir, sample), "rb")
    shutil.copyfileobj(image_file, gzip.open(f"{target_image_dir}/{prefix}_{serial_number}_000.nii.gz", "wb"))
    shutil.copyfileobj(label, gzip.open(f"{target_label_dir}/{prefix}_{serial_number}.nii.gz", "wb"))


def convert(path: str, subdir: str = "DatasetName"):
    # INPUT DATA
    path = join(path, subdir)
    file_suffix = ".nii"

    # Train/Test Splits
    labels_dir_tr = join(path, "Labels", "Train")
    images_dir_tr = join(path, "Images", "Train")
    training_samples = subfiles(labels_dir_tr, join=False, suffix=file_suffix)

    labels_dir_ts = join(path, "Labels", "Test")
    images_dir_ts = join(path, "Images", "Test")
    test_samples = subfiles(labels_dir_ts, join=False, suffix=file_suffix)

    # OUTPUT DATA
    # Target names
    task_name = "Task000_MyTask"
    prefix = "MyTask"

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    ###Populate Target Directory###
    # This is likely also the place to apply any re-orientation, resampling and/or label correction.

    tr_cases = [
        (sample, images_dir_tr, labels_dir_tr, target_imagesTr, target_labelsTr, prefix, file_suffix)
        for sample in training_samples
    ]

    ts_cases = [
        (sample, images_dir_ts, labels_dir_ts, target_imagesTs, target_labelsTs, prefix, file_suffix)
        for sample in test_samples
    ]

    all_cases = tr_cases + ts_cases

    p = Pool(8)
    p.starmap(convert_case, all_cases)
    p.close()
    p.join()

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        ("T1.. or maybe CT?",),
        labels={0: "background", 1: "Fake Label", 2: "Also Fake Label"},
        dataset_name=task_name,
        license="hands off!",
        dataset_description="Fake Dataset",
        dataset_reference="",
    )



from batchgenerators.utilities.file_and_folder_operations import join, isdir, subdirs, maybe_mkdir_p
from dataclasses import dataclass
from typing import Union
from yucca.paths import yucca_models, yucca_preprocessed_data
from yucca.training.configuration.configure_task import TaskConfig


@dataclass
class PathConfig:
    plans_path: str
    save_dir: str
    task_dir: str
    train_data_dir: str
    version_dir: str
    version: int

    def lm_hparams(self):
        return {
            "plans_path": self.plans_path,
            "save_dir": self.save_dir,
            "train_data_dir": self.train_data_dir,
            "version_dir": self.version_dir,
            "version": self.version,
        }


def get_path_config(task_config: TaskConfig):
    task_dir = join(yucca_preprocessed_data, task_config.task)
    train_data_dir = join(task_dir, task_config.planner_name)
    save_dir = join(
        yucca_models,
        task_config.task,
        task_config.model_name + "__" + task_config.model_dimensions,
        task_config.manager_name + "__" + task_config.planner_name,
        task_config.experiment,
        f"{task_config.split_method}_{task_config.split_param}_fold_{task_config.split_idx}",
    )

    version = detect_version(save_dir, task_config.continue_from_most_recent)
    version_dir = join(save_dir, f"version_{version}")
    maybe_mkdir_p(version_dir)
    plans_path = join(task_dir, task_config.planner_name, task_config.planner_name + "_plans.json")

    return PathConfig(
        plans_path=plans_path,
        save_dir=save_dir,
        task_dir=task_dir,
        train_data_dir=train_data_dir,
        version_dir=version_dir,
        version=version,
    )


def detect_version(save_dir, continue_from_most_recent) -> Union[None, int]:
    # If the dir doesn't exist we return version 0
    if not isdir(save_dir):
        return 0

    # The dir exists. Check if any previous version exists in dir.
    previous_versions = subdirs(save_dir, join=False)
    # If no previous version exists we return version 0
    if not previous_versions:
        return 0

    # If previous version(s) exists we can either (1) continue from the newest or
    # (2) create the next version
    if previous_versions:
        newest_version = max([int(i.split("_")[-1]) for i in previous_versions])
        if continue_from_most_recent:
            return newest_version
        else:
            return newest_version + 1



# Source: https://github.com/VainF/DeepLabV3Plus-Pytorch
import torch
from torch import nn
from torch.nn import functional as F


class ASPPConv(nn.Sequential):
    def __init__(self, in_channels, out_channels, dilation):
        modules = [
            nn.Conv2d(
                in_channels,
                out_channels,
                3,
                padding=dilation,
                dilation=dilation,
                bias=False,
            ),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        ]
        super(ASPPConv, self).__init__(*modules)


class ASPPPooling(nn.Sequential):
    def __init__(self, in_channels, out_channels):
        super(ASPPPooling, self).__init__(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(in_channels, out_channels, 1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        size = x.shape[-2:]
        x = super(ASPPPooling, self).forward(x)
        return F.interpolate(x, size=size, mode="bilinear", align_corners=False)


class ASPP(nn.Module):
    def __init__(self, in_channels, atrous_rates):
        super(ASPP, self).__init__()
        out_channels = 256
        modules = []
        modules.append(
            nn.Sequential(
                nn.Conv2d(in_channels, out_channels, 1, bias=False),
                nn.BatchNorm2d(out_channels),
                nn.ReLU(inplace=True),
            )
        )

        rate1, rate2, rate3 = tuple(atrous_rates)
        modules.append(ASPPConv(in_channels, out_channels, rate1))
        modules.append(ASPPConv(in_channels, out_channels, rate2))
        modules.append(ASPPConv(in_channels, out_channels, rate3))
        modules.append(ASPPPooling(in_channels, out_channels))

        self.convs = nn.ModuleList(modules)

        self.project = nn.Sequential(
            nn.Conv2d(5 * out_channels, out_channels, 1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Dropout(0.1),
        )

    def forward(self, x):
        res = []
        for conv in self.convs:
            res.append(conv(x))
        res = torch.cat(res, dim=1)
        return self.project(res)


class DeepLabHeadV3Plus(nn.Module):
    def __init__(self, in_channels, low_level_channels, num_classes, aspp_dilate=[12, 24, 36]):
        super(DeepLabHeadV3Plus, self).__init__()
        self.project = nn.Sequential(
            nn.Conv2d(low_level_channels, 48, 1, bias=False),
            nn.BatchNorm2d(48),
            nn.ReLU(inplace=True),
        )

        self.aspp = ASPP(in_channels, aspp_dilate)

        self.classifier = nn.Sequential(
            nn.Conv2d(304, 256, 3, padding=1, bias=False),
            nn.BatchNorm2d(256),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, num_classes, 1),
        )
        self._init_weight()

    def forward(self, feature):
        low_level_feature = self.project(feature["low_level"])
        output_feature = self.aspp(feature["out"])
        output_feature = F.interpolate(
            output_feature,
            size=low_level_feature.shape[2:],
            mode="bilinear",
            align_corners=False,
        )
        return self.classifier(torch.cat([low_level_feature, output_feature], dim=1))

    def _init_weight(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec 16 13:25:09 2022

@author: zcr545
"""
from copy import deepcopy
import numpy as np


def get_bbox_for_foreground(array, background_label=0):
    array = deepcopy(array)
    array[array != background_label] = 1
    return get_bbox_for_label(array, label=1)


def get_bbox_for_label(array, label, padding=0):
    a = np.where(array == label)

    x1, x2, y1, y2 = np.min(a[0]), np.max(a[0]), np.min(a[1]), np.max(a[1])

    xmin = min(x1, x2) - padding
    xmax = max(x1, x2) + padding

    ymin = min(y1, y2) - padding
    ymax = max(y1, y2) + padding

    if len(array.shape) == 3:
        z1, z2 = np.min(a[2]), np.max(a[2])

        zmin = min(z1, z2) - padding
        zmax = max(z1, z2) + padding

        return [xmin, xmax + 1, ymin, ymax + 1, zmin, zmax + 1]

    return [xmin, xmax + 1, ymin, ymax + 1]


class Box(object):
    def __init__(self, array, label, padding):
        # Create bounding box for the object.
        self.array = array
        self.label = label
        self.padding = padding
        self.set_coordinates()

    def set_coordinates(self):
        box = get_bbox_for_label(self.array, self.label, self.padding)

        assert len(box) in [4, 6], "invalid box dimensions. Should be " f"4 (2D) or 6 (3D) but is {len(box)}"
        if len(box) == 6:
            self.xmin, self.xmax, self.ymin, self.ymax, self.zmin, self.zmax = box
        if len(box) == 4:
            self.xmin, self.xmax, self.ymin, self.ymax = box

    @property
    def area(self):
        """
        Calculates the surface area. useful for IOU!
        """
        width = abs(self.xmax - self.xmin)
        height = abs(self.ymax - self.ymin)
        depth = abs(self.zmax - self.zmin)
        return width * height * depth

    def intersect(self, bbox):
        # Checks that exclude intersection
        if self.xmin > bbox.xmax or bbox.xmin > self.xmax:
            return False

        if self.ymin > bbox.ymax or bbox.ymin > self.ymax:
            return False

        if self.zmin > bbox.zmax or bbox.zmin > self.zmax:
            return False

        x1 = min(self.xmax, bbox.xmax)
        x2 = max(self.xmin, bbox.xmin)

        y1 = min(self.ymax, bbox.ymax)
        y2 = max(self.ymin, bbox.ymin)

        z1 = min(self.zmax, bbox.zmax)
        z2 = max(self.zmin, bbox.zmin)

        intersection = max(x1 - x2, 0) * max(y1 - y2, 0) * max(z1 - z2, 0)
        return intersection

    def iou(self, bbox):
        intersection = self.intersect(bbox)
        if not intersection:
            return False

        iou = intersection / float(self.area + bbox.area - intersection)
        # return the intersection over union value
        return iou



import torch
from fvcore.nn import FlopCountAnalysis


def maybe_to_gpu(data):
    device = get_available_device()

    if isinstance(data, list):
        return [d.to(device, non_blocking=True) for d in data]
    return data.to(device, non_blocking=True)


def get_available_device():
    if torch.cuda.is_available():
        return "cuda"
    elif torch.backends.mps.is_available():
        return "mps"
    else:
        return "cpu"


def flush_and_get_torch_memory_allocated(device):
    if isinstance(device, torch.device):
        device = device.type  # Get the name: str of the torch.device such as "cuda", "mps" or "cpu"
    if device == "cuda":
        torch.cuda.empty_cache()
        return torch.cuda.memory_allocated(device)
    elif device == "mps":
        torch.mps.empty_cache()
        return torch.mps.driver_allocated_memory()
    else:
        return 0


def measure_FLOPs(model: torch.nn.Module, data: torch.Tensor):
    # Returns a FlopCountAnalysis object
    # Use obj.total() for total # FLOPs
    # Use obj.by_module() for # FLOPs pr. module
    return FlopCountAnalysis(model, data)



"""
Implementation from: https://colab.research.google.com/drive/1bzCH3Yaq8gK0ZByxlcaRaj3pOc2u6zht#scrollTo=ezUlfYJ59jWl
Described in: https://medium.com/deep-learning-for-protein-design/a-comprehensive-guide-to-memory-usage-in-pytorch-b9b7c78031d3

Formula (from above source):

Let m = model memory

Let f = the amount of memory consumed by the forward pass for a batch_size of 1.

Let g = m be the amount of memory for the gradients.

Let d = 1 if training on one GPU and 2 if training on >1 GPU.

Let o = the number of moments stored by the optimizer (probably 0, 1, or 2)

Let b = 0.5 if using mixed precision training, and 1 if using full precision training.

Then for training,

Max memory consumption = m + f*batch_size*b + d*g + o*m
"""

import torch
import numpy as np
import yucca
import math
import logging
from yucca.utils.torch_utils import get_available_device, flush_and_get_torch_memory_allocated
from yucca.utils.files_and_folders import recursive_find_python_class
from yucca.utils.kwargs import filter_kwargs

from batchgenerators.utilities.file_and_folder_operations import join
from torch import nn


def estimate_memory_training(model, model_input, optimizer_type=torch.optim.Adam, use_amp=True, device=None):
    """Predict the maximum memory usage of the model.
    Args:
        optimizer_type (Type): the class name of the optimizer to instantiate
        model (nn.Module): the neural network model
        sample_input (torch.Tensor): A sample input to the network. It should be
            a single item, not a batch, and it will be replicated batch_size times.
        batch_size (int): the batch size
        use_amp (bool): whether to estimate based on using mixed precision
        device (torch.device): the device to use
    """
    device = torch.device(get_available_device())
    # Reset model and optimizer
    model.cpu()
    optimizer = optimizer_type(model.parameters(), lr=0.001)

    a = flush_and_get_torch_memory_allocated(device)
    model.to(device)
    b = flush_and_get_torch_memory_allocated(device)
    model_memory = b - a
    _ = model(model_input.to(device)).sum()
    c = flush_and_get_torch_memory_allocated(device)
    if use_amp:
        amp_multiplier = 0.5
    else:
        amp_multiplier = 1
    forward_pass_memory = (c - b) * amp_multiplier
    gradient_memory = model_memory
    if isinstance(optimizer, torch.optim.Adam):
        o = 2
    elif isinstance(optimizer, torch.optim.RMSprop):
        o = 1
    elif isinstance(optimizer, torch.optim.SGD):
        o = 0
    elif isinstance(optimizer, torch.optim.Adagrad):
        o = 1
    else:
        raise ValueError(
            "Unsupported optimizer. Look up how many moments are"
            + "stored by your optimizer and add a case to the optimizer checker."
        )
    gradient_moment_memory = o * gradient_memory
    total_memory_bytes = model_memory + forward_pass_memory + gradient_memory + gradient_moment_memory
    total_memory_gb = total_memory_bytes * 1e-9
    return total_memory_gb


def find_optimal_tensor_dims(
    dimensionality,
    num_classes,
    modalities,
    model_name,
    max_patch_size,
    fixed_patch_size: tuple | list = None,
    fixed_batch_size: tuple | list = None,
    max_memory_usage_in_gb=None,
):
    if max_memory_usage_in_gb is None:
        try:
            gpu_vram_in_gb = int(torch.cuda.get_device_properties(0).total_memory / 1024**2 * 0.001)
        except RuntimeError:
            gpu_vram_in_gb = 12
        # Don't wanna utilize more than 8/12GB, to ensure epoch times are kept relatively low
        if dimensionality == "2D":
            max_memory_usage_in_gb = min(8, gpu_vram_in_gb)
        if dimensionality == "3D":
            max_memory_usage_in_gb = min(12, gpu_vram_in_gb)

    # Use this offset to factor the overhead from CUDA and other libraries taking a substantial amount of VRAM
    offset = 2.5

    OOM_OR_MAXED = False
    final_batch_size = None
    final_patch_size = None

    if dimensionality == "2D":
        if len(max_patch_size) == 3:
            max_patch_size = max_patch_size[1:]
        conv = nn.Conv2d
        dropout = nn.Dropout2d
        norm = nn.InstanceNorm2d
        batch_size = 8
        max_batch_size = 512
        patch_size = [32, 32]
        absolute_max = 384**2
    if dimensionality == "3D":
        conv = nn.Conv3d
        dropout = nn.Dropout3d
        norm = nn.InstanceNorm3d
        batch_size = 2
        max_batch_size = 2
        patch_size = [32, 32, 32]
        absolute_max = 128**3

    if fixed_batch_size:
        batch_size = fixed_batch_size
        max_batch_size = fixed_batch_size

    if fixed_patch_size is not None:
        patch_size = fixed_patch_size
        # first fix dimensions so they are divisible by 16 (otherwise issues with standard pools and strides)
        patch_size = [math.ceil(i / 16) * 16 for i in patch_size]
        max_patch_size = patch_size
        if fixed_batch_size:  # In this case we just instantly return after dims are fixed
            return batch_size, tuple(patch_size)

    model = recursive_find_python_class(
        folder=[join(yucca.__path__[0], "network_architectures")],
        class_name=model_name,
        current_module="yucca.network_architectures",
    )
    model_kwargs = {
        "input_channels": modalities,
        "num_classes": num_classes,
        "conv_op": conv,
        "patch_size": patch_size,
        "dropout_op": dropout,
        "norm_op": norm,
    }
    model_kwargs = filter_kwargs(model, model_kwargs)
    model = model(**model_kwargs)

    est = 0
    idx = 0
    maxed_idxs = []

    while not OOM_OR_MAXED:
        try:
            if np.prod(patch_size) >= absolute_max:
                max_patch_size = patch_size

            inp = torch.zeros((batch_size, modalities, *patch_size))
            est = estimate_memory_training(model, inp)

            # If estimated usage is still within acceptable bounds we set the (maybe temporary) final dimensions
            if est < max_memory_usage_in_gb - offset:
                final_batch_size = batch_size
                final_patch_size = tuple(patch_size)
            else:
                OOM_OR_MAXED = True

            if patch_size[idx] + 16 < max_patch_size[idx]:
                patch_size[idx] += 16
                if idx < len(patch_size) - 1:
                    idx += 1
                else:
                    idx = 0
            else:
                # here we mark that one dimension has been maxed out
                if idx not in maxed_idxs:
                    maxed_idxs.append(idx)
                # if not all dimensions are maxed out for the patch_size,
                # we try the next dimension
                if not len(maxed_idxs) == len(patch_size):
                    if idx < len(patch_size) - 1:
                        idx += 1
                    else:
                        idx = 0

            # when all dimensions of the patch are maxed
            # we try increasing the batch_size instead
            if len(maxed_idxs) == len(patch_size):
                # Unless batch_size is maxed
                if not max_batch_size > batch_size:
                    final_batch_size = batch_size
                    final_patch_size = tuple(patch_size)
                    OOM_OR_MAXED = True
                if len(patch_size) == 3:
                    batch_size += 2
                else:
                    batch_size += 4
        except torch.cuda.OutOfMemoryError:
            OOM_OR_MAXED = True
    if final_batch_size is None or final_patch_size is None:
        logging.warn(
            "\n"
            "Final batch and/or patch size was not found. \n"
            "This is likely caused by supplying large fixed parameters causing (or almost causing) OOM errors. \n"
            "Will attempt to run with supplied parameters, but this might cause issues."
        )
        logging.warn(
            f"Estimated GPU memory usage for parameters is: {est}GB and the max requested vram is: {max_memory_usage_in_gb-offset}GB. \n"
            f"This includes an offset of {offset}GB to account for vram used by PyTorch and CUDA. \n"
            "Consider increasing the max vram or working with a smaller batch and/or patch size."
            "\n"
        )
        if final_batch_size is None:
            final_batch_size = batch_size
        if final_patch_size is None:
            final_patch_size = tuple(patch_size)

    logging.info(
        f"Final batch and patch sizes set to {final_batch_size} and {final_patch_size} based on the current constraints. \n"
        f"Max GPU memory usage: {max_memory_usage_in_gb}GB. \n"
        f"Estimated GPU memory usage: {est+offset}. This includes an offset of {offset}GB to allow a margin of error and account for VRAM grabbed by torch and other backend libraries. \n"
        f"Max patch size: {max_patch_size} \n"
        f"Max batch size: {max_batch_size} \n"
    )
    del model, inp
    return final_batch_size, final_patch_size


if __name__ == "__main__":
    batch_size, patch_size = find_optimal_tensor_dims(
        dimensionality="2D",
        max_memory_usage_in_gb=12,
        num_classes=3,
        modalities=1,
        model_name="MedNeXt",
        max_patch_size=[191, 167],
    )



"""
This script can be used in two ways:
1. You can specify the paths to the folder containing predictions and labels/ground truth
using "--pred" and "--gt" and define the labels of interest using e.g. "-c 0 1"

2. You can specify the task, trainer and planner like it's also done in other yucca_ scripts.
if "-t" for target task is left blank, we assume you are predicting and evaluating data
from the same task as the one the model is trained on.
"""

import argparse
from yucca.utils.saving import merge_softmax_from_folders


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--in_dirs",
        help="",
        required=True,
        nargs="*",
    )
    parser.add_argument(
        "--out_dir",
        help="Name of the target task i.e. the data to be predicted. " "Should be of format: TaskXXX_MYTASK",
        required=False,
    )
    args = parser.parse_args()

    merge_softmax_from_folders(folders=args.in_dirs, outpath=args.out_dir)


if __name__ == "__main__":
    main()



from yucca.paths import yucca_raw_data
from batchgenerators.utilities.file_and_folder_operations import subdirs


def maybe_get_task_from_task_id(task_id: str | int):
    task_id = str(task_id)
    tasks = subdirs(yucca_raw_data, join=False)

    # Check if name is already complete
    if task_id in tasks:
        return task_id

    # If not, we try to recreate the name
    # We use the raw_data folder as reference
    for task in tasks:
        if task_id.lower() in task.lower():
            return task

    # If we can't find anything we just return the original, on the offchance that the task does not exist in Raw Data while existing in e.g. Preprocessed
    print(
        f"Couldn't find a task called: {task_id} in the raw data folder: {yucca_raw_data}. If your task only exists in e.g. the Preprocessed folder things might still work."
    )
    return task_id



import numpy as np
import numpy.typing as npt
from typing import Literal
from sklearn.metrics import roc_auc_score


def dice(tp, fp, tn, fn):  # noqa: U100
    try:
        return (2 * tp) / (2 * tp + fp + fn)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def dice_per_label(tp_list, fp_list, tn_list, fn_list):
    return [dice(tp_list[i], fp_list[i], tn_list[i], fn_list[i]) for i in range(len(tp_list))]


def jaccard(tp, fp, tn, fn):  # noqa: U100
    try:
        return (tp) / (tp + fp + fn)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def jaccard_per_label(tp_list, fp_list, tn_list, fn_list):
    return [jaccard(tp_list[i], fp_list[i], tn_list[i], fn_list[i]) for i in range(len(tp_list))]


def sensitivity(tp, fp, tn, fn):  # noqa: U100
    # recall, hit rate, tpr
    # How many cases of X are correctly recognized as X?
    try:
        return tp / (tp + fn)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def specificity(tp, fp, tn, fn):  # noqa: U100
    # beware with using TN-metrics.
    try:
        return tn / (tn + fp)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def precision(tp, fp, tn, fn):  # noqa: U100
    # When X is predicted, how often is it truly X?
    try:
        return tp / (tp + fp)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def volume_similarity(tp, fp, tn, fn):  # noqa: U100
    try:
        return 1 - abs(fn - fp) / (2 * tp + fn + fp)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def f1(tp, fp, tn, fn):  # noqa: U100
    try:
        prec = precision(tp, fp, tn, fn)
        sens = sensitivity(tp, fp, tn, fn)
        return 2 * (prec * sens) / (prec + sens)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def accuracy(tp, fp, tn, fn):
    try:
        return (tp + tn) / (tp + fp + tn + fn)
    except (ZeroDivisionError, RuntimeWarning):
        if tp + fn > 0:
            return 0
        else:
            return np.nan


def auroc(y_true: npt.ArrayLike, y_score: npt.ArrayLike, multi_class_mode: Literal["ovr", "ovo"] = "ovr"):
    """
    Compute Area Under the Receiver Operating Characteristic Curve (ROC AUC/AUROC) from prediction scores for each class.

    :param y_true: ground truth labels, shape (n_samples,n_classes) for multiclass or (n_samples,) for binary.
                    Both probabilities and labels are accepted.
    :param y_score: predicted scores, shape (n_samples,n_classes)
    :param multi_class_mode: In multiclass setting, calculate ROC AUC one-vs-one (ovo), or one-vs-rest (ovr). One of {'ovr', 'ovo'}, default='ovr'

    """
    assert len(y_true.shape) == 1 or y_true.shape == y_score.shape, "y_true must be 1D or 2D"
    return roc_auc_score(y_true, y_score, average=None, multi_class=multi_class_mode)


def TP(tp, fp, tn, fn):  # noqa: U100
    return tp


def FP(tp, fp, tn, fn):  # noqa: U100
    return fp


def FN(tp, fp, tn, fn):  # noqa: U100
    return fn


def total_pos_gt(tp, fp, tn, fn):  # noqa: U100
    return tp + fn


def total_pos_pred(tp, fp, tn, fn):  # noqa: U100
    return tp + fp



import argparse
import yucca
from yucca.utils.task_ids import maybe_get_task_from_task_id
from yucca.utils.files_and_folders import recursive_find_python_class
from batchgenerators.utilities.file_and_folder_operations import join


def main():
    parser = argparse.ArgumentParser()

    # Required Arguments #
    parser.add_argument(
        "-t",
        "--task",
        help="Name of the task used for training. "
        "The data should already be preprocessed using yucca_preprocess"
        "Argument should be of format: TaskXXX_MYTASK",
    )
    parser.add_argument("-chk", "--checkpoint", help="Path to the checkpoint from where the weights should be restored. ")

    # Optional arguments with default values #
    parser.add_argument(
        "-d",
        help="Dimensionality of the Model. Can be 3D or 2D. "
        "Defaults to 3D. Note that this will always be 2D if ensemble is enabled.",
        default="3D",
    )
    parser.add_argument(
        "-m",
        help="Model Architecture. Should be one of MultiResUNet or UNet"
        " Note that this is case sensitive. "
        "Defaults to the standard UNet.",
        default="UNet",
    )
    parser.add_argument(
        "-man",
        help="Manager Class to be used. " "Defaults to the basic YuccaManager",
        default="YuccaManager",
    )
    parser.add_argument(
        "-pl",
        help="Plan ID to be used. "
        "This specifies which plan and preprocessed data to use for training "
        "on the given task. Defaults to the YuccaPlanne folder",
        default="YuccaPlanner",
    )

    # Optionals that can be changed experimentally. For long term solutions these should be specified by a unique Manager.
    parser.add_argument(
        "--batch_size",
        type=int,
        default=None,
        help="Batch size to be used for training. Overrides the batch size specified in the plan.",
    )
    parser.add_argument("--disable_logging", help="disable logging.", action="store_true", default=False)
    parser.add_argument("--ds", help="Used to enable deep supervision", default=False, action="store_true")
    parser.add_argument(
        "--epochs", help="Used to specify the number of epochs for training. Default is 1000", type=int, default=1000
    )
    parser.add_argument("--experiment", help="A name for the experiment being performed, with no spaces.", default="finetune")
    parser.add_argument("--loss", help="Should only be used to employ alternative Loss Function", default=None)
    parser.add_argument(
        "--lr",
        help="Should only be used to employ alternative Learning Rate. Format should be scientific notation e.g. 1e-4.",
        default=1e-3,
    )
    parser.add_argument("--max_vram", type=int, default=12)
    parser.add_argument("--mom", help="Should only be used to employ alternative Momentum.", default=0.9)
    parser.add_argument(
        "--new_version",
        help="Start a new version, instead of continuing from the most recent. ",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--num_workers",
        type=int,
        help="Num workers used in the DataLoaders. By default this will be inferred from the number of available CPUs-1",
        default=None,
    )
    parser.add_argument(
        "--patch_size",
        nargs="+",
        type=str,
        help="Use your own patch_size. Example: if 32 is provided and the model is 3D we will use patch size (32, 32, 32). This patch size can be set manually by passing 32 32 32 as arguments. The argument can also be min, max or mean.",
        default=None,
    )
    parser.add_argument("--precision", type=str, default="bf16-mixed")
    parser.add_argument("--profile", help="Enable profiling.", action="store_true", default=False)
    parser.add_argument("--split_idx", type=int, help="idx of splits to use for training.", default=0)
    parser.add_argument(
        "--split_data_method", help="Specify splitting method. Either kfold, simple_train_val_split", default="kfold"
    )
    parser.add_argument(
        "--split_data_param",
        help="Specify the parameter for the selected split method. For KFold use an int, for simple_split use a float between 0.0-1.0.",
        default=5,
    )
    parser.add_argument("--train_batches_per_step", type=int, default=250)
    parser.add_argument("--val_batches_per_step", type=int, default=50)

    args = parser.parse_args()

    task = maybe_get_task_from_task_id(args.task)
    checkpoint = args.checkpoint
    dimensions = args.d
    model_name = args.m
    manager_name = args.man
    planner = args.pl

    batch_size = args.batch_size
    log = not args.disable_logging
    deep_supervision = args.ds
    epochs = args.epochs
    experiment = args.experiment
    loss = args.loss
    lr = args.lr
    max_vram = args.max_vram
    momentum = args.mom
    new_version = args.new_version
    num_workers = args.num_workers
    patch_size = args.patch_size
    precision = args.precision
    profile = args.profile
    split_idx = args.split_idx
    split_data_method = args.split_data_method
    split_data_param = args.split_data_param
    train_batches_per_step = args.train_batches_per_step
    val_batches_per_step = args.val_batches_per_step

    kwargs = {}

    manager = recursive_find_python_class(
        folder=[join(yucca.__path__[0], "training", "managers")],
        class_name=manager_name,
        current_module="yucca.training.managers",
    )

    manager = manager(
        batch_size=batch_size,
        ckpt_path=checkpoint,
        continue_from_most_recent=not new_version,
        deep_supervision=deep_supervision,
        enable_logging=log,
        experiment=experiment,
        loss=loss,
        learning_rate=lr,
        max_epochs=epochs,
        max_vram=max_vram,
        model_dimensions=dimensions,
        model_name=model_name,
        momentum=momentum,
        num_workers=num_workers,
        patch_size=patch_size,
        planner=planner,
        precision=precision,
        profile=profile,
        split_idx=split_idx,
        split_data_method=split_data_method,
        split_data_param=split_data_param,
        step_logging=False,
        task=task,
        train_batches_per_step=train_batches_per_step,
        val_batches_per_step=val_batches_per_step,
        **kwargs,
    )
    manager.run_finetuning()


if __name__ == "__main__":
    main()



import numpy as np


def get_steps_for_sliding_window(shape: np.array, patch_size: tuple, overlap: int):
    assert overlap >= 0 and overlap < 1, "make sure overlap is 0 =< overlap < 1"
    steplist = []

    for i, patch in enumerate(patch_size):
        # To compute the steps for an image with dimensions=(251, 214, 198)
        # and patch size=(64, 128, 96):
        # First get all the possible steps between 0 and 251
        # with step size=64 * 1-overlap.
        # i.e. overlap = 0 means we step a full patch each time,
        slices = slice(0, shape[i], np.floor(patch * (1 - overlap)).astype(int))
        steps = np.arange(shape[i])[slices]

        # Then we get the final step start-idx and remove any steps starting
        # at a higher idx than this.
        # For a dimension of 251 pixels/voxels with a patch size of 64
        # the last start-idx will be 251-64=187.
        final_step = shape[i] - patch
        steps = steps[steps < final_step]
        steps = np.append(steps, final_step)

        steplist.append(steps)
    return steplist



"""
Task_conversion file for the Fetal Planes DB (Barcelona) dataset (https://doi.org/10.5281/zenodo.3904280).

This is an example of a classification dataset with 2D PNG images and image-level labels.
"""

import numpy as np
import pandas as pd
import shutil
from tqdm import tqdm
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subdirs, subfiles
from yucca.paths import yucca_raw_data
from yucca.task_conversion.utils import generate_dataset_json


def convert(path: str, subdir: str = "tiny-imagenet-200"):
    # INPUT DATA

    path = f"{path}/{subdir}"
    # file_suffix = ".png"

    # Train and Val (We use Val as Test) images are in the same folder
    train_images_dir = join(path, "train")
    test_images_dir = join(path, "val", "images")
    # OUTPUT DATA
    # Target names
    task_name = "Task502_tiny_imagenet_200"
    prefix = "tiny_imagenet_200"

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    # collect labels
    next_label = 1
    labeldict = {}

    val_annotations = pd.read_csv(
        join(path, "val", "val_annotations.txt"),
        sep="\t",
        header=None,
        usecols=[0, 1],
    )

    train_categories = subdirs(train_images_dir, join=False)
    for category in tqdm(train_categories, total=len(train_categories)):
        if labeldict.get(category) is None:
            labeldict[category] = next_label
            next_label += 1
        label = np.array([labeldict[category]], dtype=np.uint8)
        for image in subfiles(join(train_images_dir, category, "images"), join=False):
            serial_number = image[: -len(".JPEG")]
            image_file_path = join(train_images_dir, category, "images", image)
            shutil.copyfile(image_file_path, f"{target_imagesTr}/{prefix}_{serial_number}_000.png")
            np.savetxt(f"{target_labelsTr}/{prefix}_{serial_number}.txt", label, fmt="%d")

    test_images = subfiles(test_images_dir, join=False)
    for image in tqdm(test_images, total=len(test_images)):
        category = val_annotations[1].iloc[np.where(val_annotations[0] == image)].values[0]
        assert labeldict.get(category) is not None
        label = np.array([labeldict[category]], dtype=np.uint8)
        serial_number = image[: -len(".JPEG")]
        image_file_path = join(test_images_dir, image)
        shutil.copyfile(image_file_path, f"{target_imagesTs}/{prefix}_{serial_number}_000.png")
        np.savetxt(f"{target_labelsTs}/{prefix}_{serial_number}.txt", label, fmt="%d")

    # Now construct the pair of label-ids and real labels

    label_names = pd.read_csv(
        join(path, "words.txt"),
        sep="\t",
        header=None,
        usecols=[0, 1],
    )
    labeldict_with_words = {}
    for key, val in labeldict.items():
        word = label_names[1].iloc[np.where(label_names[0] == key)].values[0]
        labeldict_with_words[val] = key + " = " + word
    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        ("Natural Images",),
        labels=labeldict_with_words,
        dataset_name=task_name,
        license="CC-BY 4.0",
        dataset_description="Tiny ImageNet 200",
        dataset_reference="",
    )



# %%
from yucca.training.managers.YuccaManager import YuccaManager


class YuccaManager_NoPatches(YuccaManager):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(
            patch_based_training=False,
            *args,
            **kwargs,
        )


# %%
